import sys

from rest_helpers import responses, flask, framework_adapter

# Hack needed until we move to python3.7
sys.modules[__name__] = framework_adapter.Proxy(responses, flask.flask_adapter_builder)




import sys
from rest_helpers import routes, aiohttp, framework_adapter
# Hack needed until we move to python3.7
sys.modules[__name__] = framework_adapter.Proxy(routes, aiohttp.aiohttp_adapter_builder)





from mock import patch, Mock, MagicMock

from rest_helpers.jsonapi_objects import Resource
from rest_helpers.framework_adapter import BaseFrameworkAdapter


class TestRequestContext(object):#pragma: no cover
    def __init__(self, app, **options):
        self.app = app
        self.options = options
        self.test_ctx = None

    def __enter__(self):
        self.test_ctx = self.app.test_request_context(**self.options)
        self.test_ctx.push()

    def __exit__(self, type, value, traceback):
        self.test_ctx.pop()

class TestClass(Resource):#pragma: no cover
    resource_type = "/tests"
    def __init__(self, id = 1):
        super(TestClass, self).__init__("test_name", "test_type")
        self.id=id
        self.list = ["1",2,"3"]
        self.dic = {
            "entry_1": "value 1",
            "list": [{"x":1,"b":"2"},{"x":2,"b":"3"},{"x":3,"b":"4"}]
        }

class SubTestClass(Resource): #pragma: no cover
    resource_type = "/tests/subtests"
    def __init__(self, id = 1):
        super(SubTestClass, self).__init__("sub_test_name" )
        self.id=id
        self.list = ["1",2,"3"]
        self.dic = {
            "entry_1": "value 1",
            "list": [{"x":1,"b":"2"},{"x":2,"b":"3"},{"x":3,"b":"4"}]
        }


import functools

from schematics.models import Model
from schematics import types
from schematics.exceptions import ValidationError


def _schematic_model_validator(model, post=False):
    if not post:
        #schematics models are validated post deserialization
        return True,""
    else:
        try:
            model.validate()
            return True,""
        except ValidationError as ex:
            return False, str(ex.messages)

def _schematic_type_validator(schematic_type, value, post=False):
    if not post:
        #schematics types are validated post deserialization
        return True,""
    else:
        try:
            schematic_type.validate(value)
            return True,""
        except ValidationError as ex:
            return False, str(ex.messages)

# the order is important here because we will use isinstance, so we use a list
type_to_validator_tuple_list = list({
    Model: (_schematic_model_validator,),
    types.BaseType: (lambda type_hint: functools.partial(_schematic_type_validator, type_hint),)
}.items())

def get_type_validators(type_hint):
    potential_type_to_validator_tuple = next((x for x in type_to_validator_tuple_list
                                            if isinstance(type_hint, type) and issubclass(type_hint, x[0]) or isinstance(type_hint, x[0])), None)
    if potential_type_to_validator_tuple:
        return potential_type_to_validator_tuple[1][0](type_hint) if isinstance(potential_type_to_validator_tuple[1], tuple) else potential_type_to_validator_tuple[1]
    else:
        return None


import sys
import pytest
import asyncio
import httpretty

from mock import patch, Mock, MagicMock
from rest_helpers import binding, validators, framework_adapter, rest_exceptions
from rest_helpers.tests import test_common

@pytest.fixture
def test_adapter(request):
    adapter = MagicMock(wraps=framework_adapter.BaseFrameworkAdapter())
    adapter.is_in_test = lambda:False
    adapter.get_rest_helper_request_context = lambda:None
    adapter.make_json_response = lambda  obj,status=200, headers=None: (obj, status, headers)
    return adapter

@pytest.mark.asyncio
async def test_bind_hints(counter, test_adapter):
    binding_decorator_1 = binding.base_binder(test_adapter)
    binding_decorator_1.get_value = asyncio.coroutine(lambda:"1")

    binding_decorator_2 = binding.base_binder(test_adapter)
    binding_decorator_2.get_value = asyncio.coroutine(lambda:"true")

    @binding.bind_hints(test_adapter)
    def inner_func(
        abc: (int, binding_decorator_1),
        cde: (bool, binding_decorator_2)):
        counter["inner_func"] += 1
        assert abc == 1
        assert cde == True
        return "success"

    response = await inner_func()

    # This will verify that the function is not called multiple times
    assert counter["inner_func"] == 1

    # This will verify that the assertion in the inner function succeeded
    assert response == "success"

@pytest.mark.asyncio
async def test_bind_hints_with_no_hints(counter, test_adapter):
    """
    This use to cause an infinite recursion when the wrapper was
    added even when no bind hints was used
    """

    @binding.bind_hints(test_adapter)
    def inner_func(
        abc,
        cde):
        counter["inner_func"] += 1
        assert abc == 1
        assert cde == True
        return "success"

    response = inner_func(1, True)

    # This will verify that the function is not called multiple times
    assert counter["inner_func"] == 1

    # This will verify that the assertion in the inner function succeeded
    assert response == "success"

#region base_binder
@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_with_default_values(counter, test_adapter):
    binding_decorator = binding.base_binder(test_adapter, field="abc")
    binding_decorator.get_value = asyncio.coroutine(lambda:"value")

    @binding_decorator
    def inner_func(abc):
        counter["inner_func"] += 1
        assert abc == "value"
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_with_deserializer(counter, test_adapter):
    binding_decorator = binding.base_binder(test_adapter, field="abc", deserializer=lambda x:test_obj(x))
    binding_decorator.get_value = asyncio.coroutine(lambda:"value")

    class test_obj:
        def __init__(self, dic):
            self.dic = dic

    @binding_decorator
    def inner_func(abc):
        counter["inner_func"] += 1
        assert isinstance(abc, test_obj)
        assert abc.dic == "value"
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"


@pytest.mark.asyncio
@pytest.mark.base_binder
@pytest.mark.parametrize("exception,expected", [
    (binding.MissingFieldException("This is a missing field exception."), binding.MissingFieldException),
    (binding.InvalidDataException("This is a invalid data exception."), rest_exceptions.InvalidDataException),
    (Exception("This is a server exception."), Exception),
])
async def test_base_binder_with_exception(counter, test_adapter, exception, expected):
    binding_decorator = binding.base_binder(test_adapter, field="abc")
    binding_decorator.get_value = asyncio.coroutine(lambda:(_ for _ in ()).throw(exception))

    @binding_decorator
    def inner_func(abc): #pragma: no cover
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"] == 0
    assert isinstance(excep, expected)

@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_with_default_value_field(counter, test_adapter):
    binding_decorator = binding.base_binder(test_adapter, field="abc")
    binding_decorator.get_value = asyncio.coroutine(lambda:(_ for _ in ()).throw(binding.MissingFieldException))

    @binding_decorator
    def inner_func(abc = None):
        counter["inner_func"] += 1
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"


@pytest.mark.asyncio
@pytest.mark.base_binder
@pytest.mark.parametrize("data,expected,inner_func_count", [
    ("valid", "success",1),
    ("not valid", "field1 must have value 1", 0),
])
async def test_base_binder_with_validator(counter, test_adapter, data, expected, inner_func_count):
    def validator(data, post):
        counter["validator"] += 1
        return (data=="valid", "field1 must have value 1.")

    binding_decorator = binding.base_binder(test_adapter, field="abc", validator=validator)
    binding_decorator.get_value = asyncio.coroutine(lambda:data)

    @binding_decorator
    def inner_func(abc):
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"]==inner_func_count
    assert counter["validator"]>=1
    assert expected in str(excep if inner_func_count == 0 else response)


@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_in_test_mode(counter, test_adapter):
    test_adapter.is_in_test=lambda:True

    binding_decorator = binding.base_binder(test_adapter, field="abc")

    @binding_decorator
    def inner_func(abc):
        counter["inner_func"] += 1
        assert abc == 'value1'
        return "success"

    response = await inner_func('value1')
    assert counter["inner_func"] == 1
    assert response == "success"


@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_on_async_function(counter, test_adapter):
    binding_decorator = binding.base_binder(test_adapter, field="abc")
    binding_decorator.get_value = asyncio.coroutine(lambda:"value")

    @binding_decorator
    async def inner_func(abc):
        counter["inner_func"] += 1
        assert abc == 'value'
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_with_improper_field(counter, test_adapter):
    binding_decorator = binding.base_binder(test_adapter, field="foo")
    with pytest.raises(Exception) as ex:
        @binding_decorator
        def inner_func(abc): #pragma: no cover
            counter["inner_func"] += 1
            assert abc == "value"
            return "success"

    assert "Argument foo could not be found in function" in str(ex.value)

async def test_base_binder_default_validator(test_adapter, make_coro):
    with patch("rest_helpers.validators.get_type_validators", MagicMock(return_value= lambda x,y: (1,""))) as get_type_validators:
        binding_decorator = binding.base_binder(test_adapter, field="abc")
        binding_decorator.get_value = lambda : make_coro("1")
        @binding_decorator
        def inner_func(abc:int):
            return "success" #pragma: no cover
        await inner_func()
        get_type_validators._mock_call_args[0][0] == int

        binding_decorator = binding.base_binder(test_adapter, field="abc", type=bool)
        binding_decorator.get_value = lambda : make_coro("True")
        @binding_decorator
        def inner_func(abc):
            return "success" #pragma: no cover
        await inner_func()
        get_type_validators._mock_call_args[0][0] == bool


@pytest.mark.base_binder
async def test_base_binder_default_deserializer(test_adapter, make_coro):
    with patch("rest_helpers.type_deserializers.get_default_deserializer", MagicMock(return_value= lambda x: 1)) as get_default_deserializer:
        binding_decorator = binding.base_binder(test_adapter, field="abc")
        binding_decorator.get_value = lambda : make_coro("1")
        @binding_decorator
        def inner_func(abc:int):
            return "success" #pragma: no cover
        await inner_func()
        get_default_deserializer._mock_call_args[0][0] == int

        binding_decorator = binding.base_binder(test_adapter, field="abc", type=bool)
        binding_decorator.get_value = lambda : make_coro("1")
        @binding_decorator
        def inner_func(abc):
            return "success" #pragma: no cover
        await inner_func()
        get_default_deserializer._mock_call_args[0][0] == bool

@pytest.mark.asyncio
@pytest.mark.base_binder
async def test_base_binder_stacking(test_adapter, counter):
    binding_decorator_abc = binding.base_binder(test_adapter, field="abc")
    binding_decorator_abc.get_value = asyncio.coroutine(lambda:"valueABC")

    binding_decorator_cde = binding.base_binder(test_adapter, field="cde")
    binding_decorator_cde.get_value = asyncio.coroutine(lambda:"valueCDE")

    @binding_decorator_abc
    @binding_decorator_cde
    def inner_func(abc, cde):
        counter["inner_func"] += 1
        assert abc == "valueABC"
        assert cde == "valueCDE"
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

    assert binding_decorator_abc.real_view_function.__name__ == "inner_func"
    assert binding_decorator_cde.real_view_function.__name__ == "inner_func"
#endregion

#region from_json_body
@pytest.mark.asyncio
async def test_from_json_body_with_default_values(counter, test_adapter):
    test_adapter.get_current_request_body = asyncio.coroutine(lambda:'{"field1":"value1", "field2":"value2"}')
    @binding.from_json_body(test_adapter)
    def inner_func(data):
        counter["inner_func"] += 1
        assert data == { 'field1':'value1', 'field2':'value2'}
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"


@pytest.mark.asyncio
async def test_from_json_body_with_validator_missing_data(counter, test_adapter):
    test_adapter.get_current_request_body = asyncio.coroutine(lambda:'')

    @binding.from_json_body(test_adapter)
    def inner_func(data): #pragma: no cover
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"] == 0
    assert "The body of the request is not valid: a json object is expected." in str(excep)

@pytest.mark.asyncio
@pytest.mark.parametrize("version, expected", [
    ("v1", {'a': 1}),
    ("v2", {'a': 'b', 'field1': 'value1', 'field2': 'value2'})
])
async def test_from_json_body_with_versionner(test_adapter, version, expected):
    test_adapter.get_current_request_body = asyncio.coroutine(lambda:'{"field1":"value1", "field2":"value2"}')

    from rest_helpers.versioning import UrlRootVersionner
    class TestVersionner(UrlRootVersionner):
        class v1:
            def body(self, body):
                return '{   "a": 1 }'

        class v2:
            def body_dict(self, body):
                body["a"] = "b"
                return body

        class value3:
            pass

    rh_context = MagicMock()
    test_adapter.get_rest_helper_request_context = lambda :rh_context
    rh_context.versionner = TestVersionner()

    @binding.from_json_body(test_adapter)
    def inner_func(data):
        assert isinstance(data, dict)
        return data

    rh_context.versionner.requested_version = version
    response = await inner_func()
    assert response == expected


#endregion

#region field_from_json_body

@pytest.mark.asyncio
@pytest.mark.field_from_json_body
@pytest.mark.parametrize("field, json_field, data", [
    ("fieldA", "field1", '{"field1":"value1", "field2":"value2"}'),
    ('field1/field2/fieldA', None,'{"field1":{"field2":{"fieldA":"value1"}}, "field2":"value2"}')
])
async def test_field_from_json_body_with_specific_json_field(counter, test_adapter, field, json_field, data):
    test_adapter.get_current_request_body = asyncio.coroutine(lambda:data)

    @binding.field_from_json_body(test_adapter, field=field, json_field=json_field)
    def inner_func(fieldA):
        counter["inner_func"] += 1
        assert fieldA == 'value1'
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.field_from_json_body
@pytest.mark.parametrize("data,expected", [
    ('{"field_a":"b"', "request data is not valid JSON"),
    ('{"field_b":"b_value"}', "The field field_a is not present in the content of the request."),
])
async def test_field_from_json_body_invalid_body(counter, test_adapter, data, expected):
    test_adapter.get_current_request_body = asyncio.coroutine(lambda:data)

    @binding.field_from_json_body(test_adapter, "field_a")
    def inner_func(field_a): #pragma: no cover
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"] == 0
    assert expected in str(excep)

#endregion

#region field_from_query_string
@pytest.mark.asyncio
@pytest.mark.field_from_query_string
async def test_field_from_query_string_with_specific_query_field(counter, test_adapter):
    test_adapter.get_current_request_query_string_args = lambda:{"fieldA":["value1"], "field2":["value2"]}

    @binding.from_query_string(test_adapter, field="field1", query_field="fieldA")
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == 'value1'
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.field_from_query_string
async def test_field_from_query_string_with_list(counter, test_adapter):
    test_adapter.get_current_request_query_string_args = lambda:{"field1":["value1", "value3"], "field2":["value2"]}

    @binding.from_query_string(test_adapter, field="field1", as_list=True)
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == ["value1", "value3"]
        return "success"

    response = await inner_func()
    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.field_from_query_string
async def test_field_from_query_string_with_missing_field(counter, test_adapter):
    test_adapter.get_current_request_query_string_args = lambda:{"field1":["value1"], "field2":["value2"]}

    @binding.from_query_string(test_adapter, field="fieldX")
    def inner_func(fieldX): #pragma: no cover
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"]==0
    assert "The field fieldX is not present in the query string." in str(excep)

@pytest.mark.asyncio
async def test_from_query_string_with_versionner(test_adapter):
    test_adapter.get_current_request_query_string_args = lambda:{"field1":["value1"], "field2":["value2"]}

    from rest_helpers.versioning import UrlRootVersionner
    class TestVersionner(UrlRootVersionner):
        class v1:
            def query_string_args(self, query_string_args):
                query_string_args["a"]="b"
                return query_string_args

        class v2:
            pass

    rh_context = MagicMock()
    test_adapter.get_rest_helper_request_context = lambda :rh_context
    rh_context.versionner = TestVersionner()

    @binding.from_query_string(test_adapter, field="a")
    def inner_func(a):
        return a

    rh_context.versionner.requested_version = "v1"
    response = await inner_func()
    assert response == "b"

#endregion

#region field_from_header

@pytest.mark.asyncio
@pytest.mark.field_from_header
async def test_field_from_header_with_specific_field(counter, test_adapter):
    test_adapter.get_current_request_headers_dict = lambda:{"fieldA":"value1", "field2":["value2"]}

    @binding.from_header(test_adapter, field="field1", header_field="fieldA")
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == 'value1'
        return "success"

    response = await inner_func()

    assert counter["inner_func"] == 1
    assert response == "success"

@pytest.mark.asyncio
@pytest.mark.field_from_header
async def test_field_from_header_with_missing_field(counter, test_adapter):
    test_adapter.get_current_request_headers_dict = lambda:{"field1":["value1"], "field2":["value2"]}

    @binding.from_header(test_adapter, field="fieldX")
    def inner_func(fieldX): #pragma: no cover
        counter["inner_func"] += 1
        return "success"

    excep = None
    try:
        response = await inner_func()
    except Exception as ex:
        excep = ex

    assert counter["inner_func"]==0
    assert "The field fieldX is not present in the requests headers." in str(excep)

@pytest.mark.asyncio
async def test_from_header_with_versionner(test_adapter):
    test_adapter.get_current_request_headers_dict = lambda:{"field1":["value1"], "field2":["value2"]}

    from rest_helpers.versioning import UrlRootVersionner
    class TestVersionner(UrlRootVersionner):
        class v1:
            def headers(self, headers):
                headers["a"]="b"
                return headers

        class v2:
            pass

    rh_context = MagicMock()
    test_adapter.get_rest_helper_request_context = lambda :rh_context
    rh_context.versionner = TestVersionner()

    @binding.from_header(test_adapter, field="a")
    def inner_func(a):
        return a

    rh_context.versionner.requested_version = "v1"
    response = await inner_func()
    assert response == "b"
#endregion


#region oauth

@pytest.mark.asyncio
@pytest.mark.oauth
@pytest.mark.parametrize("get_unverified_header, get_unverified_claims, allowed_domains, expected_exception, decode_side_effect", [
    ({},{}, ["A.com","B.C.org"], rest_exceptions.UnauthorizedException, Exception()),
    ({"kid":"ab129f_Z"}, {"iss": "http://d.ca/my/endpoint"}, ["A.com","B.C.org"], rest_exceptions.UnauthorizedException, Exception()),
    ({"kid":"ab129f_Z"}, {"iss": "http://A.com/my/endpoint"}, ["A.com","B.C.org"], rest_exceptions.UnauthorizedException, Exception()),
    ({"kid":"abc_1"}, {"iss": "http://A.com/my/endpoint"}, None, rest_exceptions.ForbiddenException, Exception()),
    ({"kid":"abc_1"}, {"iss": "http://A.com/my/endpoint"}, ["A.com","B.C.org"], None, lambda *x,**kw: {"A":1, "B":{"C":2}})
])
async def test_oauth_with_invalid_token(counter, test_adapter, get_unverified_header, get_unverified_claims, allowed_domains, expected_exception, decode_side_effect):
    test_adapter.get_current_request_headers_dict = lambda:{"Authorization":"ABC"}

    @binding.from_Oauth(test_adapter, field="field1", allowed_domains=allowed_domains, validate_options={"a":1, "b":2}, client_id="TheClientId")
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == {"A":1, "B":{"C":2}}
        return "success"

    with patch("jose.jws.get_unverified_header") as m_get_unverified_header:
        m_get_unverified_header.return_value = get_unverified_header
        with patch("jose.jwt.get_unverified_claims") as m_get_unverified_claims:
            with patch("jose.jwt.decode") as m_decode:
                m_decode.side_effect = decode_side_effect
                m_get_unverified_claims.return_value = get_unverified_claims
                httpretty.enable()
                httpretty.register_uri(
                    httpretty.GET,
                    "http://A.com/my/endpoint/.well-known/openid-configuration",
                    body='{"jwks_uri": "http://B.com/a/b/c"}'
                )
                httpretty.register_uri(
                    httpretty.GET,
                    "http://B.com/a/b/c",
                    body='{"keys": [ {"kid":"abc-1"}, {"kid":"abc-2"} ]}'
                )

                excep = None
                response = None
                try:
                    response = await inner_func()
                except Exception as ex:
                    excep = ex
                httpretty.disable()

                if response ==  "success":
                    assert counter["inner_func"] == 1
                else:
                    assert isinstance(excep, expected_exception)
                    assert counter["inner_func"] == 0

@pytest.mark.asyncio
@pytest.mark.oauth
async def test_oauth_cache(counter, test_adapter):
    test_adapter.get_current_request_headers_dict = lambda:{"Authorization":"ABC"}

    decorator = binding.from_Oauth(test_adapter, field="field1", allowed_domains=["A.com","B.C.org"], validate_options={"a":1, "b":2}, client_id="TheClientId")
    decorator._public_keys = {"abc1": "abc"}

    @decorator
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == {"A":1, "B":{"C":2}}
        return "success"

    with patch("jose.jws.get_unverified_header") as m_get_unverified_header:
        m_get_unverified_header.return_value = {"kid":"abc_1"}
        with patch("jose.jwt.decode") as m_decode:
            m_decode.return_value = {"A":1, "B":{"C":2}}

            response = await inner_func()

            assert response ==  "success"
            assert counter["inner_func"] == 1

@pytest.mark.asyncio
@pytest.mark.Oauth
async def test_Oauth_hardcoded_valid_tokens(counter, test_adapter):
    test_adapter.get_current_request_headers_dict = lambda:{"Authorization":"ABC"}

    decorator = binding.from_Oauth(test_adapter, field="field1", valid_tokens={"ABC":{"A":1, "B":{"C":2}}})
    decorator._public_keys = {"abc1": "abc"}

    @decorator
    def inner_func(field1):
        counter["inner_func"] += 1
        assert field1 == {"A":1, "B":{"C":2}}
        return "success"

    response = await inner_func()

    assert response ==  "success"
    assert counter["inner_func"] == 1

#endregion


import functools
from decimal import Decimal
import datetime
from dateutil import parser

from schematics.models import Model
from schematics import types

def bool_deserializer(x):
    """
    This is to be used to deserialize boolean : we consider that if
    the value is "", then we should return true in order to be
    able to parse boolean from query strings.
    eg: http://myhost.com/index?loggedin&admin&page=5
    is equivalent to
        http://myhost.com/index?loggedin=True&admin=True&page=5
    """
    return x is not None and (x == "" or x.lower() == "true" )

def int_deserializer(x):
    """
    This is to be used to deserialize integers.
    """
    return int(x)

def float_deserializer(x): # pragma: no cover (nothing to test here)
    """
    This is to be used to deserialize floats.
    """
    return float(x)

def decimal_deserializer(x):
    """
    This is to be used to deserialize decimals.
    """
    return Decimal(x)

def datetime_deserializer(x): # pragma: no cover (nothing to test here)
    """
    This is to be used to deserialize datetimes.
    """
    return parser.parse(x)

def string_deserializer(x): # pragma: no cover (nothing to test here)
    """
    This is to be used to deserialize strings.
    """
    return str(x)

# the order is important here because we will use isinstance, so we use a list
type_to_deserializer_tuple_list = list({
    bool: bool_deserializer, # bool must be first because a bool is an int
    int: int_deserializer,
    float: float_deserializer,
    Decimal: decimal_deserializer,
    str: string_deserializer,
    datetime.datetime: datetime_deserializer,
    Model: (lambda type_hint: type_hint,),
    types.BaseType: (lambda type_hint: type_hint.to_native,)
}.items())

def get_default_deserializer(type_hint):
    potential_deserializer_tuple = next((x for x in type_to_deserializer_tuple_list
                                            if isinstance(type_hint, type) and issubclass(type_hint, x[0]) or isinstance(type_hint, x[0])), None)
    if potential_deserializer_tuple is not None:
        return potential_deserializer_tuple[1][0](type_hint) if isinstance(potential_deserializer_tuple[1], tuple) else potential_deserializer_tuple[1]
    else:
        return None


import json
import re
import traceback
import logging

from functools import reduce
from time import time
from rest_helpers.jsonapi_objects import Error, Resource, SuccessResponse, ErrorResponse, Response, Link
from rest_helpers.type_serializers import to_jsonable,response_to_jsonable
from rest_helpers import rest_exceptions, binding

LOGGER = logging.getLogger(__name__)
def base_exception_handler(framework_adapter, exception):
    if isinstance(exception, rest_exceptions.InvalidDataException):
        return bad_request(framework_adapter, exception)
    elif isinstance(exception, binding.MissingFieldException):
        return bad_request(framework_adapter, exception)
    elif isinstance(exception, rest_exceptions.UnauthorizedException):
        return error(framework_adapter, 401, "Unauthorized", "Your authentication was not successful.")
    elif isinstance(exception, rest_exceptions.ForbiddenException):
        return error(framework_adapter, 403, "Forbidden", "You are not authorized to access the requested resources or perform the requested operation.")
    else:
        LOGGER.error(str(exception) + traceback.format_exc())
        return internal_server_error(framework_adapter, exception, traceback.format_exc())

#region error responses
def bad_request(framework_adapter, exception):
    """Creates an HTTP Status code 400 response with the given message."""
    return error(framework_adapter, 400, "Bad request", "Client error: "+str(exception))

def not_found(framework_adapter, details=None):
    """Creates an HTTP Status code 404 response with the given message."""
    return error(framework_adapter, 404, "Not found", "Client error: The requested resource does not exist.{}".format(("\n"+details) if details else ""))

def internal_server_error(framework_adapter, exception, stack_trace=""):
    """Creates an HTTP Status code 500 response with the given message."""
    details = "Server error: {exception_type} {exception_str} \n {stack_trace}".format(
        exception_type=exception.__class__.__name__,
        exception_str=str(exception),
        stack_trace=stack_trace
    )
    return error(framework_adapter, 500, "Internal server error", details)

def error(framework_adapter, status_code, title, detail, retry_after=None):
    """Creates an error response with the given status code, error title and detail."""
    assert status_code >= 400
    error_obj = Error(
        error_id = framework_adapter.get_current_request_headers_dict().get("X-Unique-ID") or str(time()),
        status = status_code,
        title = title,
        detail = detail
    )

    return _response_from_error(framework_adapter, error_obj, headers={"Retry-After": retry_after} if retry_after else None)
#end region

#region success responses
def ok(framework_adapter, data, page_size=None, id_only=False):
    """Create an HTTP response with exit code 200:OK

    Arguments:
        data {dict} -- the data to be returned

    Keyword Arguments:
        page_size {int} -- if specified, pagination will be used,
                            and this will be the number of items on
                            a page(default: {None})

    Returns:
        [dict] -- A json serializable representation of a JSONAPI response.
    """
    return success(framework_adapter, data, 200, page_size = page_size, id_only=id_only)


def created(framework_adapter, data):
    """ Create an HTTP response with exit code 201:Created """
    return success(framework_adapter, data, 201)

def accepted(framework_adapter, data=None):
    """
    Create an HTTP response with exit code 202:Accepted. It adds a meta field output
    indicating the operation was successful.
    """
    return success(framework_adapter, data,202, {"output":"success"})

def success(framework_adapter, data, status_code, meta=None, links=None, page_size=None, id_only=False):
    """ Create a success response with the given object, status code, and meta. """
    assert status_code < 300 and status_code >= 200
    request_args=framework_adapter.get_current_request_query_string_args()

    if isinstance(data, list):
        actual_page_size = request_args.get("page_size") or page_size or framework_adapter.get_rest_helper_request_context().page_size

        if actual_page_size is not None:
            meta = meta if meta is not None else {}
            links = links if links is not None else {}
            data = _paginate(framework_adapter, data, actual_page_size, links, meta)

    if data is not None and not isinstance(data, Resource) and not isinstance(data, list):
        jsonable = to_jsonable(data)
    else:
        resp = SuccessResponse(data=data, meta=meta, links=links)
        jsonable = response_to_jsonable(resp, id_only=id_only)


    if "json_path" in request_args:
        try:
            jsonable["data"] = _filter_using_json_path(framework_adapter, jsonable, request_args.get("json_path")[0].strip(" /").split("/"))
        except Exception as ex:
            return bad_request(framework_adapter, ex)

    rh_context = framework_adapter.get_rest_helper_request_context()
    jsonable = rh_context.versionner.response_body_dict(jsonable) if rh_context.versionner is not None else jsonable
    response = framework_adapter.make_json_response(jsonable, status_code)

    return response
#endregion


#region private helpers

def _paginate(framework_adapter, data, page_size, links, meta):
    assert isinstance(links, dict)
    assert isinstance(meta, dict)

    page_size = page_size if page_size is not None else \
                framework_adapter.get_rest_helper_request_context.page_size

    if callable(page_size):
        page_size = page_size()

    if page_size is None:
        return

    page_size = int(page_size)
    if "total_pages" not in meta:
        meta["total_pages"] = (len(data)+page_size-1)//page_size

    assert page_size > 0
    prev_link = None
    next_link = None

    # This is a zero based page index
    query_string_args = framework_adapter.get_current_request_query_string_args()
    page = int(query_string_args.get("page",[1])[0])-1
    left = page * page_size
    right = min(len(data), left + page_size)

    if right != len(data):
        next_link = _get_paginated_self_link(framework_adapter, page+2)
    if left != 0:
        prev_link = _get_paginated_self_link(framework_adapter, page)

    data = data[left:right]

    if prev_link is not None:
            links["last"] = prev_link

    if next_link is not None:
        links["next"] = next_link

    return data

def _get_paginated_self_link(framework_adapter, page):
    # we proceed like this to avoid changing the order of the query string args
    query_string = "&".join(x if not x.startswith("page=") else "page={0}".format(page) for x in framework_adapter.get_current_request_query_string().decode().split("&"))
    if "page=" not in query_string:
        query_string = ("{0}{1}page={2}").format(query_string, "&" if query_string != "" else "", page)

    return Link("{0}?{1}".format( framework_adapter.get_current_request_path(), query_string))

def _filter_using_json_path(framework_adapter, jsonable_data, segments):
    for i,segment in enumerate(segments):
        if isinstance(jsonable_data, list):
            if segment[0] == "*":
                filter = _parse_filter(segment)
                sub_segments = segments[i+1:] if i < len(segments)-1 else []
                jsonable_data = [_filter_using_json_path(framework_adapter, x,sub_segments) for x in jsonable_data if filter(x) is True]
                return jsonable_data

            try:
                index = int(segment)
                if index >= len(jsonable_data):
                    raise Exception("The json path provided is not valid: the index '{0}' should be lower than '{1}'.".format(segment, len(jsonable_data)))
                jsonable_data = jsonable_data[index]
            except:
                raise Exception("The json path provided is not valid: expexting an integer for list index, found '{0}'.".format(segment))
        else:
            if segment not in jsonable_data:
                raise Exception("The json path provided is not valid: the key '{0}' is not present.".format(segment))
            jsonable_data = jsonable_data[segment]

    return jsonable_data

def _parse_filter(segment):
    """This parses strings to convert them into a lambda. For instance:
    *:/key1/key2~=[^x]+x

    [
        { key1: {key2: 123x},
        { key1: {key2: 123s}
    ]
    would be filtered to [{ key1: {key2: 123x}]

    
    Arguments:
        segment {str} -- the segment to filter
    
    Raises:
        Exception -- if the filter is not properly formatted, an exception will be raised.

    Returns:
        [callable] -- a callable that will take an object as a parameter and return true or false
    """

    if len(segment) == 1:
        return lambda x: True

    segment=segment[2:]
    operators = ["~=", "==", "!="]
    op_index = next((segment.index(op) for op in operators if op in segment), None)

    if op_index is None:
        raise Exception("The filter {} provided in json_path is not valid: it must contain an operator among {}".format(segment, str(operators)))

    path = segment[:op_index].strip(" >").split(">")
    operator = segment[op_index:op_index+2]
    compared_to = segment[op_index+2:]

    compare = (lambda x,y: re.match(y,x) is not None) if operator == "~=" \
                else (lambda x,y: x == y) if operator == "==" \
                else (lambda x,y: x != y)

    # Here we use reduce drill through the object following the path.
    return lambda x: compare(str(reduce(lambda acc,cur: acc[cur], path, x)), compared_to)


def _response_from_error(framework_adapter, error, headers=None):
    error_response = ErrorResponse(error)
    response = framework_adapter.make_json_response(response_to_jsonable(error_response), error.status, headers, title = error.title)
    return response

#endregion


import os
import logging
import traceback
import functools
import inspect
from jinja2 import Template
from rest_helpers import responses, swagger, rest_helper_context, binding, await_if_needed
from rest_helpers.common import decorators
from rest_helpers.framework_adapter import BaseFrameworkAdapter

LOGGER = logging.getLogger(__name__)

_swagger_routes = []

#TODO: move the url parameter logic to the binding section

class route(object):
    """
    This decorator is to be used to create a route, catching all exceptions in order to return a well formatted 500 response.
    """

    def __init__(self, framework_adapter, rule, options=None, doc=True, versionner=None, exception_handler=None):
        assert isinstance(framework_adapter, BaseFrameworkAdapter)
        self.framework_adapter = framework_adapter
        self.options = options if options else { "methods":["GET"]}
        self.doc = doc
        self.view_function = None
        self.real_view_function = None #The view function might be multiple wrapped, this is the original viewfunction
        self.rule = rule
        self.id = None
        self.versionner = versionner
        self.exception_handler = exception_handler or functools.partial(responses.base_exception_handler, self.framework_adapter)

    async def _on_request(self, *args, **kwargs):
        try:
            rh_context = rest_helper_context.RestHelperContext()
            if self.versionner is not None:
                rh_context.versionner = self.versionner()
                rh_context.versionner.set_request_args(request_args=args, request_kwargs=kwargs)

            self.framework_adapter.attach_rest_helper_request_context(rh_context)

            self._before_fn_call(args, kwargs)

            result = await await_if_needed(self._binding_functions(*args, **kwargs))
            return result if rh_context.versionner is None else rh_context.versionner.response(result)
        except Exception as ex:
            LOGGER.error("An exception {0} occured: {1}\n stacktrace: {2} \n request url: {3} \n request body: {4}".format(
                ex.__class__.__name__,
                str(ex),
                traceback.format_exc(),
                self.framework_adapter.get_current_request_full_path(),
                await self.framework_adapter.get_current_request_body()))

            return self.exception_handler(ex)

    def _before_fn_call(self, f_arg, f_kwargs):
        pass

    def __call__(self, f):
        self.view_function = f
        self.real_view_function = f

        while hasattr(self.real_view_function, "__wrapped__"):
            self.real_view_function = self.real_view_function.__wrapped__
        self.id = decorators.get_decorated_id(self.real_view_function)

        if self.versionner is not None:
            self.versionner.version_route(self)

        self.framework_adapter.add_url_rule(self, self._on_request)
        if self.doc:
            _swagger_routes.append(self)

        self._binding_functions = binding.bind_hints(self.framework_adapter)(self.view_function)
        return self._on_request

class base_resource_route(route):
    """
    This decorator is the base class to create RESTful api routes. It handles the resource id
    binding, as well as catching all exceptions in order to return a well formatted 500 response.
    """

    def __init__(self, framework_adapter, resource_class, doc=True, options=None, versionner=None, exception_handler=None):
        super(base_resource_route, self).__init__(framework_adapter, rule=None, options=options, doc=doc, versionner=versionner, exception_handler=exception_handler)
        self.resource_class = resource_class

    def _before_fn_call(self, f_arg, f_kwargs):
        if "resource_id" not in self.real_view_function.__code__.co_varnames:
            return

        # We are using the resource type incombination with the url segments to address cases where there is a base path/prefix.
        url_segments = self.framework_adapter.get_current_request_url().strip(" /").split("/")
        type_segments = self.resource_class.resource_type.strip(" /").split('/')
        type_segments_length = len(type_segments)
        f_kwargs["resource_id"] = "".join("/{0}/{1}".format(t,url_segments[-type_segments_length*2+i*2+1]) for i,t in enumerate(type_segments))

class get_resource_route(base_resource_route):
    def __init__(self, framework_adapter, resource_class, doc=True, options=None, versionner=None, exception_handler=None):
        super(get_resource_route, self).__init__(framework_adapter, resource_class, doc, options=options, versionner=versionner, exception_handler=exception_handler)
        self.rule = "".join("/{0}/<{1}_name>".format(x, x[:-3]+'y' if x[-3:] == 'ies' else (x[:-1] if x[-1] == 's' else x)) for x in resource_class.resource_type.strip(" /").split("/"))
        self.options["methods"] = ["GET"]

class get_all_resources_route(get_resource_route):
    def __init__(self, framework_adapter, resource_class, doc=True, page_size=None, options=None, versionner=None, exception_handler=None):
        super(get_all_resources_route, self).__init__(framework_adapter, resource_class, doc, options, versionner=versionner, exception_handler=exception_handler)
        self.rule = self.rule[:self.rule.rindex("/")+1]
        self.page_size = page_size

    def _before_fn_call(self, f_arg, f_kwargs):
        self.framework_adapter.get_rest_helper_request_context().page_size = self.page_size

        if "resource_id" not in self.real_view_function.__code__.co_varnames:
            return

        # We are using the resource type incombination with the url segments to address cases where there is a base path/prefix.
        url_segments = self.framework_adapter.get_current_request_url().strip(" /").split("/")[:-1]
        type_segments = self.resource_class.resource_type.strip(" /").split('/')[:-1]
        type_segments_length = len(type_segments)
        f_kwargs["resource_id"] = "".join("/{0}/{1}".format(t,url_segments[-type_segments_length*2+i*2+1]) for i,t in enumerate(type_segments))

class delete_resource_route(base_resource_route):
    def __init__(self, framework_adapter, resource_class, doc=True, options=None, versionner=None, exception_handler=None):
        super(delete_resource_route, self).__init__(framework_adapter, resource_class, doc, options=options, versionner=versionner, exception_handler=exception_handler)
        self.rule = "".join("/{0}/<{1}_name>".format(x, x[:-3]+'y' if x[-3:] == 'ies' else (x[:-1] if x[-1] == 's' else x)) for x in resource_class.resource_type.strip(" /").split("/"))
        self.options["methods"] = ["DELETE"]

class put_resource_route(get_resource_route):
    def __init__(self, framework_adapter, resource_class, doc=True, options=None, versionner=None, exception_handler=None):
        super(put_resource_route, self).__init__(framework_adapter, resource_class, doc, options=options, versionner=versionner, exception_handler=exception_handler)
        self.options["methods"] = ["PUT"]

class patch_resource_route(get_resource_route):
    def __init__(self, framework_adapter, resource_class, doc=True, options=None, versionner=None, exception_handler=None):
        super(patch_resource_route, self).__init__(framework_adapter, resource_class, doc, options=options, versionner=versionner, exception_handler=exception_handler)
        self.options["methods"] = ["PATCH"]

class operation_resource_route(get_resource_route):
    def __init__(self, framework_adapter, resource_class, operation_name, doc=True, options=None, versionner=None, exception_handler=None):
        super(operation_resource_route, self).__init__(framework_adapter, resource_class, doc, options, versionner, exception_handler)
        self.options["methods"] = ["POST"]
        self.operation_name = operation_name
        self.rule = "{0}/{1}".format(self.rule, self.operation_name)

    def _before_fn_call(self, f_arg, f_kwargs):
        if "resource_id" not in self.real_view_function.__code__.co_varnames:
            return

        # We are using the resource type incombination with the url segments to address cases where there is a base path/prefix.
        url_segments = self.framework_adapter.get_current_request_url().strip(" /").split("/")[:-1]
        type_segments = self.resource_class.resource_type.strip(" /").split('/')
        type_segments_length = len(type_segments)
        f_kwargs["resource_id"] = "".join("/{0}/{1}".format(t,url_segments[-type_segments_length*2+i*2+1]) for i,t in enumerate(type_segments))

class group_operation_resource_route(operation_resource_route):
    def __init__(self, framework_adapter, resource_class, operation_name=None, doc=True, options=None, versionner=None, exception_handler=None):
        super(group_operation_resource_route, self).__init__(framework_adapter, resource_class, operation_name, doc, options, versionner, exception_handler)
        self.rule = "{0}/{1}".format("/".join(self.rule.split("/")[:-2]), operation_name)

    def _before_fn_call(self, f_arg, f_kwargs):
        if "resource_id" not in self.real_view_function.__code__.co_varnames:
            return

        # We are using the resource type incombination with the url segments to address cases where there is a base path/prefix.
        url_segments = self.framework_adapter.get_current_request_url().strip(" /").split("/")[:-2]
        type_segments = self.resource_class.resource_type.strip(" /").split('/')[:-1]
        type_segments_length = len(type_segments)
        f_kwargs["resource_id"] = "".join("/{0}/{1}".format(t,url_segments[-type_segments_length*2+i*2+1]) for i,t in enumerate(type_segments))
#region non decorator helpers

def add_get_resource_route(func, *args, **kwargs):
    get_resource_route(*args, **kwargs)(func)


def add_get_all_resource_route(func, *args, **kwargs):
    get_all_resources_route(*args, **kwargs)(func)

def add_put_resource_route(func, *args, **kwargs):
    put_resource_route(*args, **kwargs)(func)
    
def add_patch_resource_route(func, *args, **kwargs):
    patch_resource_route(*args, **kwargs)(func)
    
def add_operation_route(func, *args, **kwargs):
    operation_resource_route(*args, **kwargs)(func)

def add_group_operation_route(func, *args, **kwargs):
    group_operation_resource_route(*args, **kwargs)(func)

#endregion


#region extras

def add_default_swagger_routes(framework_adapter, source, basepath=None, **kwargs):
    """
    This function adds two default endpoints :
    /swagger.json => swagger spec
    / => swagger ui

    Arguments:
        framework_adapter {BaseFrameworkAdapter} -- The web framework adapter used to setup the routes.
        source {dict|func} -- The dictionary or documented method containing the service documentation.
    """
    service_description = source if isinstance(source,dict) else swagger._get_swagger_part(source, swagger.SWAGGER_DOCUMENTATION_KEY)

    basePath = basepath or service_description.get("basePath","/")
    basePath = basePath + ("/" if basePath[-1]!="/" else "")
    def get_swagger_json():
        return framework_adapter.make_json_response(swagger.get_swagger_service_description(source))

    swagger_json_route = base_resource_route(framework_adapter, None, doc=False, options={ "methods": ["GET"] })
    swagger_json_route.rule = basePath + "swagger.json"
    swagger_json_route(get_swagger_json)

    swagger_ui = None
    swagger_ui_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates", "swagger-ui.html")
    with open(swagger_ui_path, "r") as f:
        swagger_ui = Template(f.read()).render(**kwargs)
    def get_swagger_ui():
        return swagger_ui
    swagger_route = base_resource_route(framework_adapter, None, doc=False, options={ "methods": ["GET"] })
    swagger_route.rule = basePath
    swagger_route(get_swagger_ui)

#endregion



import  json
import functools

class Proxy(object):
    def __init__(self, proxied, adapter_builder):
        self.proxied = proxied
        self.adapter_builder = adapter_builder

    def __getattr__(self, name):
        def _call(*args, **kwargs):
            adapter, a_args, a_kwargs = self.adapter_builder(*args, **kwargs)
            return getattr(self.proxied, name)(adapter, *a_args, **a_kwargs)
        functools.update_wrapper(_call, getattr(self.proxied, name))
        return _call

    __file__ = ""

class BaseFrameworkAdapter(object):  # pragma: no cover
    """
    Adapter pattern: the framework adapter is here to isolate
    framework specific logic from the rest helper business logic.
    The framework adapter goal is to provide a thin adapter layer
    so that rest helpers can be used with several different web
    frameworks such as flask or aiohttp.
    """

    #region mandatory
    async def get_current_request_body(self):
        raise NotImplementedError()

    def get_current_request_query_string_args(self):
        raise NotImplementedError()

    def get_current_request_headers_dict(self):
        raise NotImplementedError()

    def get_current_request_query_string(self):
        raise NotImplementedError()

    def get_current_request_path(self):
        raise NotImplementedError()

    def attach_rest_helper_request_context(self, context):
        raise NotImplementedError()

    def get_current_request_full_path(self):
        raise NotImplementedError()

    def get_current_request_url(self):
        raise NotImplementedError()

    def add_url_rule(self, route, func):
        raise NotImplementedError()

    def get_rest_helper_request_context(self):
        raise NotImplementedError()

    def get_current_request_method(self):
        raise NotImplementedError()

    def get_current_request_headers(self):
        raise NotImplementedError()

    def make_json_response(self, obj, status = 200, headers=None):
        raise NotImplementedError()
    #endregion

    #region optional
    def is_in_test(self):
        return False

    def set_request_args(self, args):
        return args

    def set_request_kwargs(self, kwargs):
        return kwargs
    #endregion


class BaseExceptionHandle(Exception):
    def __init__(self, *args):
        super(BaseExceptionHandle, self).__init__(*args)

class NotFoundException(BaseExceptionHandle):
    """
    This exceptions indicates a client error : the object
    requested or subject to the requested operation does
    not exist.
    """
    pass

class InvalidDataException(BaseExceptionHandle):
    """
    This exception indicates a client error : the
    data sent by the client are not valid.
    """
    pass

class UnauthorizedException(BaseExceptionHandle):
    """
    This exception indicates that the client could not
    be authenticated.
    """
    pass

class ForbiddenException(BaseExceptionHandle):
    """
    This exception indicates that the client is not
    authorized to access the requested resource or
    perform the requested operation.
    """
    pass



import functools
import json
import traceback
import inspect
import typing
import requests
import re
from urllib.parse import urlparse

from jose import jws,jwt

from rest_helpers import type_deserializers, validators, await_if_needed
from rest_helpers.common import decorators
from rest_helpers.rest_exceptions import InvalidDataException, UnauthorizedException, ForbiddenException

_input_decorators = {}

class MissingFieldException(Exception):
    pass

class bind_hints(object):
    def __init__(self, framework_adapter):
        self.framework_adapter = framework_adapter


    def __call__(self, func):
        self.func = func

        return_value=func
        type_hints = self.func.__annotations__
        for field,hint in type_hints.items():
            potential_binding_decorator = hint[-1] if isinstance(hint, tuple) else hint

            # If it has the __wrapped__ attribute created by functools then it is
            # a decorator type not yet instanciated (Eg: argument:binding vs argument: binding())
            if hasattr(potential_binding_decorator, "__wrapped__"):
                potential_binding_decorator = potential_binding_decorator()
            if isinstance(potential_binding_decorator, base_binder):
                potential_binding_decorator.set_field(field)
                functools.update_wrapper(potential_binding_decorator, return_value)
                return_value = potential_binding_decorator(return_value)
                
        if return_value != func:
            functools.update_wrapper(return_value, func)
        return return_value

class base_binder(object):
    def __init__(self, framework_adapter, field=None, validator=None, deserializer=None, type=None):
        if (deserializer is not None) and (type is not None):
            raise Exception("deserializer and type cannot be provided at the same time.")

        self.func = None
        self.validator = validator
        self.deserializer = deserializer
        self.type = type
        self.framework_adapter = framework_adapter
        self.set_field(field)

    def __call__(self, f):
        self.func = f
        self.real_view_function = f

        while hasattr(self.real_view_function, "__wrapped__"):
            self.real_view_function = self.real_view_function.__wrapped__

        if self.type is None:
            type_hints = self.real_view_function.__annotations__
            self.type = type_hints.get(self.field, None)
            self.type = self.type[0] if isinstance(self.type, tuple) else self.type

        f_args = self.real_view_function.__code__.co_varnames[:self.real_view_function.__code__.co_argcount]

        try:
            arg_index = f_args.index(self.field)
        except ValueError:
            raise Exception("Argument {field} could not be found in function {function} arguments".format(field=self.field, function=self.real_view_function))

        self.has_default = f_args != None and self.real_view_function.__defaults__!=None and arg_index >= len(f_args) - len(self.real_view_function.__defaults__)
        self.default = self.real_view_function.__defaults__[arg_index - (len(f_args) - len(self.real_view_function.__defaults__))] if self.has_default else None

        async def return_value(*args, **kwargs):
            args = self.framework_adapter.set_request_args(args)
            kwargs = self.framework_adapter.set_request_kwargs(kwargs)
            return await _on_request_binding(self, *args, **kwargs)

        functools.update_wrapper(return_value, f)


        self.real_view_function_id = decorators.get_decorated_id(self.real_view_function)
        if self.real_view_function_id not in _input_decorators:
            _input_decorators[self.real_view_function_id] = []

        _input_decorators[self.real_view_function_id].append(self)

        return return_value

    async def get_value(self): #pragma no cover
        raise NotImplementedError()

    def set_field(self, field):
        self.field = field

class from_json_body(base_binder):
    __name__ = "from_json_body"
    def __init__(self, framework_adapter, field="data", validator=None, deserializer=None):
        """
        This class is to be used as a decorator:
        it will fill the parameter of a method by parsing the
        content of the body of the request.
        In particular, it will assign to the method argument named <field>
        the request content, deserialized from json to a dictionary.

        Additionally, a validator can be passed to automatically check the data.
        If the validation failed, a bad request is returned.

        Deserializer will be infered from type annotation.
        Field with a default value are considered optional.

        eg:
        request content =
        {
            field1: value1
            field2: value2
        }

        @from_json_body(field="myData")
        def method(myData):
            assert myData["field1"] == "value1"
            assert myData["field2"] == "value2"

        Arguments:
            framework_adapter {BaseFrameworkAdapter} -- The adapter used to interact with the framework

        Keyword Arguments:
            field {str} -- the name of the function argument to be filled (default: {"data"})
            validator {callable} -- a function that take the input as a value and returns a
                    tuple (bool,reason) to indicate if the input is valid or not (default: {None})
            deserializer {callable} -- a function or class used to deserialize the input to an object (default: {None}).
                                        Can be set to a schematics type or model

        """

        #TODO: test validation of object with : 1. a deserialiazer that has a validate method, a deserialized object that has a validate method.

        super(from_json_body, self).__init__(framework_adapter, field, validator, deserializer)

    async def get_value(self):
        return await _get_dict_from_json_body(self.framework_adapter)

# TODO: support optimization fot stacking
class field_from_json_body(base_binder):
    __name__ = "field_from_json_body"
    def __init__(self, framework_adapter, field=None, json_field=None, validator=None, deserializer=None):
        """
        This function is to be used as a decorator:
        it will fill the parameter of a method by parsing the
        content of the body of the request.
        In particular, it will assign to the method argument named <field>
        the value of the associated <json_field> field in the request content,
        deserialized from json to a dictionary.

        If <json_field> is not specified, the value of <field> will be used.

        Additionally, a validator can be passed to automatically check the data.
        If the validation failed, a bad request is returned.

        Deserializer will be infered from type annotation.
        Field with a default value are considered optional.

        eg:
        request content =
        {
            field1: value1
            field2: value2
        }

        @field_from_json_body(field="fieldA", json_field="field1")
        @field_from_json_body(field="fieldB", json_field="field2")
        def method(fieldA, fieldB):
            assert fieldA == "value1"
            assert fieldB == "value2"

        field = the name of the function argument to be filled. If field is
                a path, the last segment of the path will be the name of the
                function argument to be filled.
                eg: data/parent/child -> child
        json_field = the name of the field in the json body,
                    if not provided, the value of field will be used
        validator = a function that take the input as a value and returns a
                    tuple (bool,reason) to indicate if the input is valid or not
        deserializer = a function used to deserialize the input to an object
        """
        self.json_field = json_field or field
        super(field_from_json_body, self).__init__(framework_adapter, field, validator, deserializer)

    def set_field(self, field):
        if field is None:
            return

        self.json_field = self.json_field or field
        self.field = field.split("/")[-1]

    async def get_value(self):
        return await _get_field_from_json_body(self.framework_adapter, self.json_field)


class from_header(base_binder):
    __name__ = "from_header"
    def __init__(self, framework_adapter,field=None, header_field=None, validator=None, deserializer=None):
        """
        This function is to be used as a decorator:
        it will fill the parameter of a method by parsing the
        content of the headers of the request.
        In particular, it will assign to the method argument named <field>
        the value of the associated <header_field> field in the request headers.

        If <header_field> is not specified, the value of <field> will be used.

        Additionally, a validator can be passed to automatically check the data.
        If the validation failed, a bad request is returned.

        Deserializer will be infered from type annotation.
        Field with a default value are considered optional.

        eg:
        request query string field1=value1&field2=value2

        @from_header(field="fieldA", header_field="field1")
        @from_header(field="fieldB", header_field="field2")
        def method(fieldA, fieldB):
            assert fieldA == "value1"
            assert fieldB == "value2"

        field = the name of the function argument to be filled
        header_field = the name of the field in the header dict,
                    if not provided, the value of field will be used
        validator = a function that take the input as a value and returns a
                    tuple (bool,reason) to indicate if the input is valid or not
        deserializer = a function used to deserialize the input to an object
        """
        self.header_field = header_field
        super(from_header, self).__init__(framework_adapter, field, validator, deserializer)


    def set_field(self, field):
        if field is None:
            return

        self.header_field = self.header_field or field
        self.field = field

    async def get_value(self):
        return _get_field_from_headers(
            self.framework_adapter,
            self.header_field)

class from_query_string(base_binder):
    __name__ = "from_query_string"
    def __init__(self, framework_adapter,field=None, query_field=None, validator=None, deserializer=None, as_list=False):
        """
        This function is to be used as a decorator:
        it will fill the parameter of a method by parsing the
        content of the query string of the request.
        In particular, it will assign to the method argument named <field>
        the value of the associated <query_field> field in the request query string,
        deserialized to a dictionary.

        If <query_field> is not specified, the value of <field> will be used.

        Additionally, a validator can be passed to automatically check the data.
        If the validation failed, a bad request is returned.

        Deserializer will be infered from type annotation.
        Field with a default value are considered optional.

        eg:
        request query string field1=value1&field2=value2

        @from_query_string(field="fieldA", query_field="field1")
        @from_query_string(field="fieldB", query_field="field2")
        def method(fieldA, fieldB):
            assert fieldA == "value1"
            assert fieldB == "value2"

        field = the name of the function argument to be filled
        query_field = the name of the field in the query string,
                    if not provided, the value of field will be used
        validator = a function that take the input as a value and returns a
                    tuple (bool,reason) to indicate if the input is valid or not
        deserializer = a function used to deserialize the input to an object
        as_list = indicate that the input should be treated as a list at all time.
                If set to False(default), the input will be treated as a list only if the list
                contains more than one element.
        """
        self.query_field = query_field
        self.as_list = as_list

        super(from_query_string, self).__init__(framework_adapter, field, validator, deserializer)

    def set_field(self, field):
        if field is None:
            return

        self.query_field = self.query_field or field
        self.field = field

    async def get_value(self):
        return _get_field_from_query_string(
            self.framework_adapter,
            self.query_field,
            self.as_list)

_key_clean_regex=re.compile('[^a-zA-Z0-9]+')
class from_Oauth(base_binder):
    __name__ = "from_Oauth"
    def __init__(self, framework_adapter, allowed_domains=None, validate_options=None, client_id=None, audience=None, field=None, valid_tokens=None, deserializer=None):
        """
        This function is to be used as a decorator:
        it will fill the parameter of a method by parsing the
        content of the Authorization header of the request, and getting the associated
        informations from the oauth auth server if the token is valid.

        Deserializer will be infered from type annotation.
        Field with a default value are considered optional.

        allowed_domains = the domain trusted to issue tokens
        validate_options = defaults = {
                'verify_signature': True,
                'verify_aud': True,
                'verify_iat': True,
                'verify_exp': True,
                'verify_nbf': True,
                'verify_iss': True,
                'verify_sub': True,
                'verify_jti': True,
                'verify_at_hash': True,
                'leeway': 0,
            }
        client_id = the client id of the application : the token audience will be verified againt it.
        field = the method argument to be filled with auth infos
        valid_tokens = a dict of tokens that are valid and bypass auth, values associated with a valid token will be passed to the field.
        deserializer = a function used to deserialize the auth result into an object.
        """
        super(from_Oauth, self).__init__(framework_adapter, (field or "user_auth"), lambda x,y:(True,"always valid"), deserializer or (lambda x:x))
        self._public_keys = {}
        self.allowed_domains = allowed_domains
        self.client_id = client_id
        self.validate_options = validate_options
        self.valid_tokens = valid_tokens or {}
        self.audience = audience


    async def get_value(self):
        token = _get_field_from_headers(
            self.framework_adapter,
            "Authorization")
        token = token.replace("Bearer ", "")

        if token in self.valid_tokens:
            return self.valid_tokens[token]

        unauthorized_exception = UnauthorizedException("""You are not authorized to access the requested resource or operation.
            Make sure the supplied token is valid and that the user associated with it has access rights""")
        id_token = token
        decoded_token = jws.get_unverified_header(id_token)
        cleaned_key_id = None
        if 'kid' not in decoded_token:
            raise UnauthorizedException("The token does not contain a key id. Is it a proper JWT token ?")

        dirty_key_id = decoded_token['kid']

        cleaned_key_id = re.sub(_key_clean_regex, '', dirty_key_id)

        if cleaned_key_id not in self._public_keys:
            unverified_claims = jwt.get_unverified_claims(id_token)
            dirty_url = urlparse(unverified_claims['iss'])

            if self.allowed_domains is not None and dirty_url.netloc not in self.allowed_domains:
                raise UnauthorizedException("The issuer of the token does not belong to the list of approved domains : " + str(self.allowed_domains))

            cleaned_issuer = dirty_url.geturl()
            oidc_discovery_url = "{}/.well-known/openid-configuration".format(cleaned_issuer)
            openid_configuration = requests.get(oidc_discovery_url).json()
            jwks_uri = openid_configuration['jwks_uri']
            jwks = requests.get(jwks_uri).json()
            self._public_keys = {re.sub(_key_clean_regex, '',key['kid']):key for key in jwks['keys']}

            if cleaned_key_id not in self._public_keys:
                raise UnauthorizedException("The public key used to sign the token is not valid.")

        try:
            decoded = jwt.decode(
                id_token,
                key=self._public_keys[cleaned_key_id],
                audience=self.audience,
                options=self.validate_options)
        except Exception as ex:
            raise ForbiddenException("""You are not authorized to access the requested resource or operation.
            Make sure the supplied token is valid and that the user associated with it has access rights.

            Exception:
            {}
            """.format(ex))

        return decoded

#region private

async def _on_request_binding(decorator, *args, **kwargs):
    # if we are not within a request context (testing for instance)
    # there is no work to be done.
    if decorator.framework_adapter.is_in_test():
        return await await_if_needed(decorator.func(*args, **kwargs))

    if decorator.deserializer is None and decorator.type is not None:
        decorator.deserializer = type_deserializers.get_default_deserializer(decorator.type)

    if decorator.validator is None and decorator.type is not None:
        decorator.validator = validators.get_type_validators(decorator.type)

    try:
        value = await decorator.get_value()
        is_valid,reason = decorator.validator(value, False) if decorator.validator is not None else (True,"")
        if not is_valid:
            raise InvalidDataException("The value of the field {field} is not valid: {reason}".format(field=decorator.field, reason=reason))

        if callable(decorator.deserializer):
            value = decorator.deserializer(value)

        # post deserialization validation
        is_valid,reason = decorator.validator(value, True) if decorator.validator is not None else (True,"")
        if not is_valid:
            raise InvalidDataException(reason)
    except MissingFieldException as ex:
        if decorator.has_default:
            value = decorator.default
        else:
            raise

    kwargs.pop(decorator.field,None)
    kwargs[decorator.field] = value

    return await await_if_needed(decorator.func(*args, **kwargs))

async def _get_dict_from_json_body(framework_adapter):
    try:
        body = await framework_adapter.get_current_request_body()
        context = framework_adapter.get_rest_helper_request_context()

        if context is not None and context.versionner is not None:
            body = context.versionner.body(body)
        if body == "":
            raise MissingFieldException("The body of the request is not valid: a json object is expected.")
        else:
            data = json.loads(body)

        data = context.versionner.body_dict(data) if context is not None and context.versionner is not None else data
        return data
    except (SyntaxError,ValueError) as ex:
        raise InvalidDataException("request data is not valid JSON: {message}".format(message=ex))

async def _get_field_from_json_body(framework_adapter, json_field):
    data = await _get_dict_from_json_body(framework_adapter)

    json_field_segments = json_field.split("/")
    for json_field_segment in json_field_segments:
        if data is None or not isinstance(data, dict) or json_field_segment not in data:
            raise MissingFieldException("The field {json_field} is not present in the content of the request.".format(json_field=json_field))
        else:
            data = data[json_field_segment]

    return data

def _get_field_from_query_string(framework_adapter, query_field, as_list):
    query_string_args = framework_adapter.get_current_request_query_string_args()

    context = framework_adapter.get_rest_helper_request_context()
    if context is not None and context.versionner is not None:
        query_string_args = context.versionner.query_string_args(query_string_args)

    if query_field not in query_string_args:
        raise MissingFieldException("The field {query_field} is not present in the query string.".format(query_field=query_field))
    else:
        associated_list = query_string_args[query_field]
        field_value = associated_list[0] if len(associated_list) == 1 and not as_list else associated_list

    return field_value

def _get_field_from_headers(framework_adapter, header_field):
    headers = framework_adapter.get_current_request_headers_dict()
    context = framework_adapter.get_rest_helper_request_context()
    if context is not None and context.versionner is not None:
        headers = context.versionner.headers(headers)

    if header_field not in headers:
        raise MissingFieldException("The field {header_field} is not present in the requests headers.".format(header_field=header_field))
    else:
        field_value = headers[header_field]

    return field_value

#endregion



class Resource(object):
    resource_type = None
    def __init__(self, resource_name, resource_type=None, relationships=None, links=None, meta=None, parent=None):
        """
        This is the base type for a REST resource compatible with jsonapi.

        A resource is identified uniquely by the field id constructed as follow:
        id=/grand_parent_type_plural/grand_parent_resource_name/parent_type_plural/parent_name/type_plural/name
        type=/grand_parent_type_plural/parent_type_plural/type_plural

        A resource can be nested under a parent resource where it makes sense.
        Eg:
        id = /authors/mtwain/books/tomsawyer
        type = /authors/books
        call: Resource("books","tomsawyer",parent=MTwainResourceObject)

        resource_type: the type of the resource NOT including the parent type. It should be plural
                        eg: books
        resource_name: the name of the resource.
                        eg: tomsawyer
        relationships: a dictionary of related relationships objects, keyed by relation kind.
        links: a dictionary of links, keyed by kind. 'self' is a special kind: links associated to 'self' should
               point to the current resource or node.
        meta: a list of metadata objects

        """
        resource_type = resource_type if resource_type is not None else self.__class__.resource_type
        assert resource_name is not None and resource_name != ""
        assert resource_type is not None and resource_type != ""
        assert parent is None or (isinstance(parent, Resource))
        assert parent is None or resource_type.startswith(parent.type)

        assert relationships is None \
            or isinstance(relationships, dict) \
               and all(isinstance(x,Relationship) for x in relationships.values())

        assert links is None \
            or isinstance(links, dict) \
               and all(isinstance(x,Link) for x in links.values())

        self.id = "{0}/{1}/{2}".format(parent.id if parent is not None else "", resource_type.split("/")[-1], resource_name)
        self.type = resource_type
        self.name = resource_name
        self.relationships = relationships
        self.links = links
        self.meta = meta
        self._parent = parent

        # automatically create the parent relationship
        if self._parent is not None and (self.relationships is None or "parent" not in self.relationships):
            if self.relationships is None:
                self.relationships = {}
            self.relationships["parent"] = Relationship(links={ "self": Link(self.id + "/relationships/parent"), "parent": Link(parent.id)}, data = self._parent)

        if self.links == None:
            self.links = dict()

        # automatically create the self link
        if "self" not in self.links:
            self.links["self"] = Link(self.id)

class Response(object):
    def __init__(self, data, error, meta, jsonapi=None, links=None, included=None):
        """
        data: the document's primary data
        errors: an array of error objects
        meta: a meta object that contains non-standard meta-information.
        jsonapi: an object describing the server's implementation
        links: a links object related to the primary data.
        included: an array of resource objects that are related to the primary data and/or each other

        The members data and errors MUST NOT coexist in the same document.

        A document MAY contain any of these top-level members:

        jsonapi: an object describing the server's implementation
        links: a links object related to the primary data.
        included: an array of resource objects that are related to the primary
                  data and/or each other (included resources).

        If a document does not contain a top-level data key, the included member
        MUST NOT be present either.
        """
        # A response can contain either data or error but not both
        assert (data is not None) != (error is not None) or data is None
        assert data is not None or included is None
        assert data is None \
                or isinstance(data, Resource) \
                or (isinstance(data, list) and all(isinstance(x, Resource) for x in data))

        assert jsonapi is None or isinstance(jsonapi, JsonApiObject)

        assert included is None \
            or isinstance(included, list) \
               and all(isinstance(x,Resource) for x in included)

        assert links is None \
            or isinstance(links, dict) \
               and all(isinstance(x,Link) for x in links.values())

        self.data = data
        self.error = error
        self.meta = meta
        self.jsonapi = jsonapi
        self.links = links
        self.included = included

class SuccessResponse(Response):
    """
    See Response documentation.
    This response is specific to successful outcome.

    Swagger extra_definition:
        response:
            type: "object"
            properties:
                data:
                    type: "object"
    """
    def __init__(self, data, meta=None, jsonapi=None, links=None, included=None):
        super(SuccessResponse,self).__init__(data=data, error=None, meta=meta, jsonapi=jsonapi, links=links, included=included)

class ErrorResponse(Response):
    """
    See Response documentation.
    This response is specific to erroneous outcome.

    Swagger extra_definition:
    error_response:
        type: "object"
        properties:
            error:
                type: "object"
                $ref: '#/definitions/error'

    error:
        type: "object"
        properties:
            id:
                type: "string"
            status:
                type: "integer"
            title:
                type: "string"
            detail:
                type: "string"
    """
    def __init__(self, error, meta=None, jsonapi=None, links=None, included=None):
        assert error is not None
        super(ErrorResponse,self).__init__(data=None, error=error, meta=meta, jsonapi=jsonapi, links=links, included=included)


class Error(object):
    def __init__(self, error_id, status, title, detail, links=None, about=None, code=None, source=None, pointer=None, parameter=None, meta=None):
        """
        error_id: a unique identifier for this particular occurrence of the problem.
        status: the HTTP status code applicable to this problem, expressed as a string value.
        title: a short, human-readable summary of the problem that SHOULD NOT change from
               occurrence to occurrence of the problem, except for purposes of localization.
        detail: a human-readable explanation specific to this occurrence of the problem.
                Like title, this field's value can be localized.

        links: a links object containing the following members:
        about: a link that leads to further details about this particular occurrence of the problem.
        code: an application-specific error code, expressed as a string value.
        source: an object containing references to the source of the error, optionally including any of the following members:
        pointer: a JSON Pointer [RFC6901] to the associated entity in the request document [e.g. "/data" for a primary data object, or "/data/attributes/title" for a specific attribute].
        parameter: a string indicating which URI query parameter caused the error.
        meta: a meta object containing non-standard meta-information about the error.
        """
        assert status >= 400

        self.id = error_id
        self.status = status
        self.title = title
        self.detail = detail

        self.links = links
        self.about = about
        self.code = code
        self.source = source
        self.pointer = pointer
        self.parameter = parameter
        self.meta = meta

class Link(object):
    """
    Where specified, a links member can be used to represent links.
    """
    def __init__(self, url, meta = None):
        assert url is not None
        self.url = url
        self.meta = meta


class JsonApiObject(object):
    """
     JSON API document MAY include information about its implementation
     under a top level jsonapi member. If present, the value of the
     jsonapi member MUST be an object (a "jsonapi object"). The jsonapi
     object MAY contain a version member whose value is a string
     indicating the highest JSON API version supported. This object MAY
     also contain a meta member, whose value is a meta object that
     contains non-standard meta-information.
    """

    def __init__(self, meta=None):
        self.version = "1.0"
        self.meta = meta

class Relationship(object):
    """
    Members of the relationships object ("relationships") represent
    references from the resource object in which it's defined to other
    resource objects.

    Relationships may be to-one or to-many.
    """
    def __init__(self, links, data, meta=None):
        assert data is not None or links is not None
        assert links is None or isinstance(links,dict) and all(isinstance(x,Link) for x in links.values())
        assert data is None or isinstance(data,Resource)

        self.links = links
        self.data = data


import decimal
from rest_helpers import type_serializers
from rest_helpers.jsonapi_objects import Resource, Response, Link, Relationship, JsonApiObject

class TestTypeSerializers(object):
    def test_to_jsonable(self):
        class TestInnerObj(object):
            def __init__(self, message):
                self.message = message
                self._private = "private"

        class TestObj(object):
            def __init__(self):
                self.name = "test_name"
                self.id = "test_id"
                self._private = "private"
                self.__super_private = "super_private"
                self.dic = {"this": "is", "a": "dictionary", "of":"strings", "and ints":2}
                self.inner_obj = TestInnerObj("This is a message")
                self.inner_obj_list = [TestInnerObj("ex 1"), TestInnerObj("ex 2"), TestInnerObj("ex 3")]
                self.none_field = None
                self.empty_field = ""
                self.decimal = decimal.Decimal(3)
                self.decimal_float = decimal.Decimal('3.4')

        test_object = TestObj()
        assert type_serializers.to_jsonable(test_object, no_empty_field=False) == {
            "name":"test_name",
            "id":"test_id",
            "dic": {"this": "is", "a": "dictionary", "of":"strings", "and ints":2},
            "inner_obj": {"message": "This is a message"},
            "inner_obj_list" : [{"message": "ex 1"}, {"message": "ex 2"}, {"message": "ex 3"}],
            'empty_field': '',
            'none_field': None,
            'decimal': 3,
            'decimal_float': '3.4'
        }

        assert type_serializers.to_jsonable(test_object, no_empty_field=True) == {
            "name":"test_name",
            "id":"test_id",
            "dic": {"this": "is", "a": "dictionary", "of":"strings", "and ints":2},
            "inner_obj": {"message": "This is a message"},
            "inner_obj_list" : [{"message": "ex 1"}, {"message": "ex 2"}, {"message": "ex 3"}],
            'decimal': 3,
            'decimal_float': '3.4'
        }

    def test_response_to_jsonable(self):
        test_resource_1 = Resource(resource_type="/test_type", resource_name="test_name_1")
        test_response_1 = Response(data = test_resource_1, error = None, meta = {"count": 1})

        assert type_serializers.response_to_jsonable(test_response_1) == {
                'data': {
                        'attributes': {
                            'name': 'test_name_1'
                        },
                        'id': '/test_type/test_name_1',
                        'type': '/test_type',
                        'links': {'self': '/test_type/test_name_1'}
                    },
                'meta': {'count': 1}
            }

        test_resource_2 = Resource(resource_type="/test_type", resource_name="test_name_2")
        test_response_2 = Response(data = [test_resource_1, test_resource_2], error = None, meta = {"count": 2})

        assert type_serializers.response_to_jsonable(test_response_2) == {
                'data': [
                            {
                                'attributes': {'name': 'test_name_1'},
                                'id': '/test_type/test_name_1',
                                'type': '/test_type',
                                'links': {'self': '/test_type/test_name_1'}
                            },
                            {
                                'attributes': {'name': 'test_name_2'},
                                'id': '/test_type/test_name_2',
                                'type': '/test_type',
                                'links': {'self': '/test_type/test_name_2'}
                            },
                    ],
                'meta': {'count': 2}
            }

        test_response_3 = Response(data = [], error = None, meta = {"count": 0})
        assert type_serializers.response_to_jsonable(test_response_3) == {
                'data': [],
                'meta': {'count': 0}
            }

    def test_resource_to_jsonable(self):
        test_related_resource = Resource(resource_type="/american_famous_writers", resource_name="19thcentury")
        test_parent_resource = Resource(
            resource_type = "/authors",
            resource_name = "marktwain",
            relationships={
                "part of": Relationship(data=test_related_resource, links = None),
                "bio": Relationship(links={"related":Link("http://host:80/bios/mtwain")}, data=None)
            },
            links={"self":Link("http://host:80/authors/marktwain"), "related": Link("http://host:80/american_famous_writers/19thcentury")},
            meta={"count":1},
            parent=None)

        test_child_resource = Resource(
            resource_type = "/authors/books",
            resource_name = "tomsawyer",
            relationships=None,
            links={ "self": Link("http://host:80/authors/marktwain/books/tomsayer")},
            meta={"count":1},
            parent=test_parent_resource)

        assert type_serializers.resource_to_jsonable(test_child_resource) == {
            'attributes': {'name': 'tomsawyer'},
            'id': '/authors/marktwain/books/tomsawyer',
            'links': {'self': 'http://host:80/authors/marktwain/books/tomsayer'},
            'meta': {'count': 1},
            'relationships': {
                'parent': {
                    'data': {'id': '/authors/marktwain', 'type': '/authors'},
                    'links': {
                        'parent': '/authors/marktwain',
                        'self': '/authors/marktwain/books/tomsawyer?json_path=/relationships/parent'
                        }
                    }
                },
            'type': '/authors/books'
            }

        assert type_serializers.resource_to_jsonable(test_parent_resource) == {
            'attributes': {'name': 'marktwain'},
            'id': '/authors/marktwain',
            'links': {
                'related': 'http://host:80/american_famous_writers/19thcentury',
                'self': 'http://host:80/authors/marktwain'
                },
            'meta': {'count': 1},
            'relationships': {
                'bio': {
                    'links': {'related': 'http://host:80/bios/mtwain', 'self': '/authors/marktwain?json_path=/relationships/bio'}
                    },
                'part of': {
                    'data': {
                        'id': '/american_famous_writers/19thcentury',
                        'type': '/american_famous_writers'
                        },
                    'links': {'self': '/authors/marktwain?json_path=/relationships/part of'}
                    }
                },
            'type': '/authors'}

    def test_links_to_jsonable(self):
        result = type_serializers.links_to_jsonable(None)
        assert result is None

        link = Link("/a/b/c", meta={"a":1,"b":2,"c":"3"})
        result = type_serializers.links_to_jsonable({"link_type": link})

        assert result == {"link_type":
            {
                "href":"/a/b/c",
                "meta":{"a":1,"b":2,"c":"3"}
            }
        }

    def test_jsonapi_object_serializer(self):
        jsonapi_object = JsonApiObject(meta=None)
        assert type_serializers.jsonapiobject_to_jsonable(jsonapi_object) == {
            "version" : "1.0"
        }

        jsonapi_object = JsonApiObject(meta={ "a" : 1, "b": "2"})
        assert type_serializers.jsonapiobject_to_jsonable(jsonapi_object) == {
            "version" : "1.0",
            "meta" : { "a" : 1, "b": "2"}
        }

    def test_relationships_to_jsonable(self):
        link = Link("/a/b/c")
        relationship = Relationship({"related":link}, data=Resource("test_name", "/test_type"))
        assert type_serializers.relationships_to_jsonable({"rel":relationship}) == {
            "rel":{
                "data":{ "type": "/test_type", "id":"/test_type/test_name"},
                "links":{
                    "related":"/a/b/c"
                }
            }
        }

        relationship = Relationship({"related":link}, data=Resource("test_name", "/test_type"))
        assert type_serializers.relationships_to_jsonable({"rel":relationship}, "/prefix", True) == {
            "rel":{
                "data":{ "type": "/test_type", "id":"/test_type/test_name"},
                "links":{
                    "related":"/a/b/c",
                    "self":"/prefix/rel"
                }
            }
        }




import pytest
from rest_helpers import swagger, routes, framework_adapter, binding, versioning
from rest_helpers.tests import test_common


def test_update_none():
    d1 = {"a": 1, "b":2}
    swagger._update(d1, None)

    assert d1 == {"a": 1, "b":2}

def test_update_simple():
    d1 = {"a": 1, "b":2}
    d2 = {"c": 3, "b":5}

    swagger._update(d1, d2)

    assert d1 == {"a": 1, "b":5, "c":3}

def test_update_nested():
    d1 = {"a": 1, "b":{"A":"I","B":"II"}}
    d2 = {"c": 3, "b":{"A":"I","B":"V"}}

    swagger._update(d1, d2)

    assert d1 == {"a": 1, "b":{"A":"I","B":"V"}, "c":3}

def test_update_by_index():
    d1 = {"a": 1, "b":[1,2,3,4,5]}
    d2 = {"c": 3, "b":{1:"A"}}

    swagger._update(d1, d2)

    assert d1 == {"a": 1, "b":[1,"A",3,4,5], "c":3}

def test_update_list():
    d1 = {"a": 1, "b":[1,2,3]}
    d2 = {"c": 3, "b":[3,4,5]}

    swagger._update(d1,d2)

    assert d1 == {"a": 1, "b":[3,4,5], "c":3}

def test_update_list_append():
    d1 = {"a": 1, "b":[1,2,3]}
    d2 = {"c": 3, "b>+":[5]}

    swagger._update(d1,d2)

    assert d1 == {"a": 1, "b":[1,2,3,5], "c":3}

def test_update_list_delete_by_elem():
    d1 = {"a": 1, "b":[1,2,3]}
    d2 = {"c": 3, "b>-":[3]}

    swagger._update(d1,d2)

    assert d1 == {"a": 1, "b":[1,2], "c":3}

def test_update_list_delete_by_index():
    d1 = {"a": 1, "b":[1,2,3]}
    d2 = {"c": 3, "b>-":["[1]"]}

    swagger._update(d1,d2)

    assert d1 == {"a": 1, "b":[1,3], "c":3}

@pytest.mark.parametrize("type, expected",
[
    (int, "integer"),
    (str, "string"),
    (bool, "boolean")
])
def test_get_parameter_type(type, expected):
    assert swagger._get_parameter_type(type) == expected

def test_get_parameters_simple_route():
    route = routes.route(framework_adapter.BaseFrameworkAdapter(), "/test_<param1>/hello_<param2>", options=None, doc=True, versionner=None, exception_handler=None)
    parameters = swagger._get_parameters(route)

    assert parameters == [
        {
            'description': '',
            'in': 'path',
            'name': 'param1',
            'required': True,
            'type': 'string'
        },
        {
            'description': '',
            'in': 'path',
            'name': 'param2',
            'required': True,
            'type': 'string'
        }
    ]

def test_get_parameters_resource_route():
    original_binding_input_decorators = binding._input_decorators
    fw_adapter = framework_adapter.BaseFrameworkAdapter()
    try:
        route = routes.get_resource_route(fw_adapter, test_common.TestClass, options=None, doc=True, versionner=versioning.UrlRootVersionner, exception_handler=None)
        route.rule = "<rest_helper_version>/" + route.rule
        binding_list = [
                binding.from_json_body(fw_adapter, field="data", validator=None, deserializer=None),
                binding.field_from_json_body(fw_adapter, field="json_field", json_field="a/b/c", validator=None, deserializer=None),
                binding.from_header(fw_adapter, field="header_field", header_field="in_header", validator=None, deserializer=None),
                binding.from_query_string(fw_adapter, field="query_field", query_field="in_query", validator=None, deserializer=None, as_list=False),
            ]
        for i,b in enumerate(binding_list):
            b.has_default = i % 2 == 0
            b.default = "X"+str(i) if i % 2 == 0 else None

        binding._input_decorators = {
            route.id: binding_list
        }
        parameters = swagger._get_parameters(route)

        assert parameters == [
                {
                    'description': 'Version of the api to use.',
                    'enum': [],
                    'in': 'path',
                    'name': 'rest_helper_version',
                    'required': True,
                    'type': 'string'
                },
                {
                    'description': 'Name of the /tests to get.',
                    'in': 'path',
                    'name': 'test_name',
                    'required': True,
                    'type': 'string'
                },
                {
                    'in': 'query',
                    'name': 'in_query',
                    'required': True,
                    'type': 'string'
                },
                {
                    'in': 'header',
                    'name': 'in_header',
                    'required': False,
                    'type': 'string'
                },
                {
                    'in': 'body',
                    'name': 'body',
                    'required': True,
                    'schema': {
                        'type': 'object'
                    }
                }
            ]
    finally:
        binding._input_decorators = original_binding_input_decorators



import pytest
import functools
from mock import patch, Mock, MagicMock
from rest_helpers import routes, framework_adapter, binding, responses, jsonapi_objects
from rest_helpers.jsonapi_objects import Resource
from rest_helpers.tests.test_common import TestRequestContext as RequestContext, TestClass, SubTestClass

class TestAdapter(framework_adapter.BaseFrameworkAdapter):
    def __init__(self):
        self.add_url_rule = MagicMock()
        self.get_current_request_full_path = MagicMock()
        self.attach_rest_helper_request_context = MagicMock()
        self.make_json_response = MagicMock(side_effect=lambda *a,**k: a)
        self.get_rest_helper_request_context = MagicMock()

    async def get_current_request_body(self):
        return ""


def _decorator(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        return f(*args, **kwargs)
    return inner

@pytest.mark.asyncio
async def test_route():
    adapter = TestAdapter()

    exception = None
    options = {"a":"b"}
    try:
        routes.route(MagicMock(), rule="/test", doc=True, options=options)
    except Exception as ex:
        exception = ex

    assert exception is not None

    class TestExc(Exception):
        pass

    route = routes.route(adapter, "/test", doc=True, options=options, exception_handler=lambda ex: "CDE" if isinstance(ex,TestExc) else None)
    assert route.rule == "/test"

    d = {"return_value": 5}

    # This decorator is here to verrify that stacking work
    @route
    @_decorator
    def test_function(data):
        if d["return_value"] == 5:
            return d["return_value"]
        elif d["return_value"] == 4:
            raise TestExc("This is a test exc")
        else:
            raise Exception("an error occured")

    # functools wrap changes the attributes of the decorator function
    # to make it look like the decorated function : in order to make sure that
    # the fields are setup properly, we need to look at the __code__ and specifically
    # the argument of the function to make sure view_function points to the
    # decorator and real_view_function points to the actual function.
    assert route.real_view_function.__code__.co_varnames == ('data',) and route.real_view_function.__code__.co_name == "test_function"
    assert len(adapter.add_url_rule._mock_call_args_list) == 1
    assert adapter.add_url_rule._mock_call_args[0][0] == route
    assert adapter.add_url_rule._mock_call_args[0][1].__code__.co_name == "_on_request"
    assert routes._swagger_routes[0] == route

    assert await test_function(None) == 5
    d["return_value"] = 4
    response = await test_function(None)
    assert response == "CDE"

    d["return_value"] = None
    route.exception_handler = MagicMock(return_value="ABC")
    response = await test_function(None)
    assert response == "ABC"


@pytest.mark.asyncio
async def test_route_with_versionner():
    from rest_helpers.versioning import UrlRootVersionner
    class TestVersionner(UrlRootVersionner):
        class v1:
            def response(self, response):
                return 2
        class v2:
            pass

    adapter = TestAdapter()
    route = routes.route(adapter, "/test", versionner=TestVersionner)


    # This decorator is here to verrify that stacking work
    @route
    @_decorator
    def test_function(x=None, abc=None):
        return {"a":1, "b":{"c":2}}

    assert route.rule == '/<rest_helper_version>/test'

    assert await test_function(rest_helper_version="v2") == {"a":1, "b":{"c":2}}
    assert await test_function(rest_helper_version="v1") == 2
    assert await test_function() == {"a":1, "b":{"c":2}} == {"a":1, "b":{"c":2}}


@pytest.mark.asyncio
async def test_base_resource_route_resource_id_binding():
    adapter = TestAdapter()
    options = {"a":"b"}
    d = {}
    route = routes.base_resource_route(adapter, TestClass, doc=True, options=options)
    @route
    @_decorator
    def test_function(resource_id):
        d["resource_id"] = resource_id
        return "Ok"

    subroute = routes.base_resource_route(adapter, SubTestClass, doc=True, options=options)
    @subroute
    @_decorator
    def subtest_function(resource_id):
        d["resource_id"] = resource_id
        return "Ok"

    adapter.get_current_request_url = MagicMock(return_value="/tests/test_name")
    await test_function()
    assert d["resource_id"] == "/tests/test_name"

    adapter.get_current_request_url = MagicMock(return_value="/tests/test_name/subtests/subtest_name")
    await subtest_function()
    assert d["resource_id"] == "/tests/test_name/subtests/subtest_name"

@pytest.mark.asyncio
@pytest.mark.parametrize("route, route_kwargs, rule, method, res_id, url",[
    (routes.get_resource_route, { "resource_class":TestClass }, "/tests/<test_name>", "GET", "/tests/test_name", "/tests/test_name"),
    (routes.get_all_resources_route, { "resource_class":TestClass }, "/tests/", "GET", "", "/tests/"),
    (routes.put_resource_route, { "resource_class":TestClass }, "/tests/<test_name>", "PUT", "/tests/test_name", "/tests/test_name"),
    (routes.patch_resource_route, { "resource_class":TestClass }, "/tests/<test_name>", "PATCH", "/tests/test_name", "/tests/test_name"),
    (routes.delete_resource_route, { "resource_class":TestClass }, "/tests/<test_name>", "DELETE", "/tests/test_name", "/tests/test_name"),
    (routes.operation_resource_route ,{ "resource_class":TestClass, "operation_name":"test_op"}, "/tests/<test_name>/test_op", "POST", "/tests/test_name", "/tests/test_name/test_op"),
    (routes.group_operation_resource_route ,{ "resource_class":TestClass, "operation_name":"test_op"}, "/tests/test_op", "POST", "", "/tests/test_op"),

    (routes.get_resource_route, { "resource_class":SubTestClass }, "/tests/<test_name>/subtests/<subtest_name>", "GET", "/tests/test_name/subtests/subtest", "/tests/test_name/subtests/subtest"),
    (routes.get_all_resources_route, { "resource_class":SubTestClass }, "/tests/<test_name>/subtests/", "GET", "/tests/test_name", "/tests/test_name/subtests/"),
    (routes.put_resource_route, { "resource_class":SubTestClass }, "/tests/<test_name>/subtests/<subtest_name>", "PUT", "/tests/test_name/subtests/subtest", "/tests/test_name/subtests/subtest"),
    (routes.patch_resource_route, { "resource_class":SubTestClass }, "/tests/<test_name>/subtests/<subtest_name>", "PATCH", "/tests/test_name/subtests/subtest", "/tests/test_name/subtests/subtest"),
    (routes.delete_resource_route, { "resource_class":SubTestClass }, "/tests/<test_name>/subtests/<subtest_name>", "DELETE", "/tests/test_name/subtests/subtest", "/tests/test_name/subtests/subtest"),
    (routes.operation_resource_route, { "resource_class":SubTestClass, "operation_name":"test_op"}, "/tests/<test_name>/subtests/<subtest_name>/test_op", "POST", "/tests/test_name/subtests/subtest", "/tests/test_name/subtests/subtest/test_op"),
    (routes.group_operation_resource_route, { "resource_class":SubTestClass, "operation_name":"test_op"}, "/tests/<test_name>/subtests/test_op", "POST", "/tests/test_name", "/tests/test_name/subtests/test_op"),
])
async def test_resource_routes(route, route_kwargs, rule, method, res_id, url):
    adapter = TestAdapter()
    d={}
    # This decorator is here to verrify that stacking work
    r = route(adapter, **route_kwargs)
    @r
    @_decorator
    def test_function(resource_id):
        d["resource_id"] = resource_id
        return "Ok"

    assert r.rule == rule
    assert r.options["methods"] == [method]
    assert adapter.add_url_rule.mock_calls[0][1][0] == r
    adapter.get_current_request_url = MagicMock(return_value=url)
    assert await test_function() == "Ok"
    assert d["resource_id"] == res_id

    @route(adapter, **route_kwargs)
    @_decorator
    def test_function2():
        return "Ok"
    assert await test_function2() == "Ok"


@pytest.mark.parametrize("helper_method, kwargs",[
    (routes.add_get_resource_route, { "resource_class": TestClass}),
    (routes.add_get_all_resource_route, { "resource_class": TestClass}),
    (routes.add_put_resource_route, { "resource_class": TestClass}),
    (routes.add_patch_resource_route, { "resource_class": TestClass}),
    (routes.add_operation_route, { "resource_class": TestClass, "operation_name":'test_op'}),
    (routes.add_group_operation_route, { "resource_class": TestClass, "operation_name":'test_group_op'}),
])
def test_add_route_helpers(helper_method, kwargs):
    adapter = TestAdapter()

    @_decorator
    def test_function(): # pragma: no cover : This is a function used within a test
        pass

    helper_method(test_function, adapter, **kwargs)
    adapter.add_url_rule.assert_called_once()
    route_dict = adapter.add_url_rule._mock_call_args[0][0].__dict__
    assert adapter.add_url_rule.call_count == 1
    for k,v in kwargs.items():
        assert route_dict[k] == v

def test_add_swagger_route():
    adapter = TestAdapter()
    routes.add_default_swagger_routes(adapter, {})
    assert adapter.add_url_rule.call_count == 2
    assert adapter.add_url_rule.mock_calls[0][1][0].rule == "/swagger.json"
    assert adapter.add_url_rule.mock_calls[1][1][0].rule == "/"

    routes.add_default_swagger_routes(adapter, {"basePath":"/v3"})
    assert adapter.add_url_rule.call_count == 4
    assert adapter.add_url_rule.mock_calls[2][1][0].rule == "/v3/swagger.json"
    assert adapter.add_url_rule.mock_calls[3][1][0].rule == "/v3/"

    routes.add_default_swagger_routes(adapter, {}, basepath="/v3")
    assert adapter.add_url_rule.call_count == 6
    assert adapter.add_url_rule.mock_calls[4][1][0].rule == "/v3/swagger.json"
    assert adapter.add_url_rule.mock_calls[5][1][0].rule == "/v3/"


#TODO : test get all resources route with paging



import inspect
from rest_helpers.rest_exceptions import InvalidDataException

class BaseVersionner:
    def __init__(self):
        attributes = (getattr(self, a) for a in dir(self))
        innner_classes = (c for c in attributes if inspect.isclass(c) and c != self.__class__)
        self.supported_versions = {c.name if hasattr(c, "name") else c.__name__:c() for c in innner_classes}
        self.requested_version = None

    @staticmethod
    def version_route(route):
        raise NotImplementedError

    def get_specific_versionner(self):
        version = self.requested_version or list(self.supported_versions.keys())[-1] #default version is the latest
        if version not in self.supported_versions:
            raise InvalidDataException("The specified version ({})is not correct: it should be among {}".format(version, str(self.supported_versions.keys())))

        return self.supported_versions[version]

    # The following methods verify if the specific versionner has the corresponding
    # method before calling it instead of using an inheritance mechanism: this is so
    # the definition of inner versionner is clear and looks very readable.
    def body(self, body):
        versionner = self.get_specific_versionner()
        return versionner.body(body) if hasattr(versionner, "body") else body

    def body_dict(self, body):
        """
        If you  want to version a json object, use this hook instead of
        Tthe body hook that would force you to deserialize the json to
        reserialize the versionned one right after

        Arguments:
            body {dict} -- the body of the request aleready deserialized from json.

        Returns:
            {dict} -- the versionned body dictionary.
        """

        versionner = self.get_specific_versionner()
        return versionner.body_dict(body) if hasattr(versionner, "body_dict") else body

    def response(self, response):
        versionner = self.get_specific_versionner()
        return versionner.response(response) if hasattr(versionner, "response") else response

    def headers(self, headers):
        versionner = self.get_specific_versionner()
        return versionner.headers(headers) if hasattr(versionner, "headers") else headers

    def query_string_args(self, query_string_args):
        versionner = self.get_specific_versionner()
        return versionner.query_string_args(query_string_args) if hasattr(versionner, "query_string_args") else query_string_args

    def response_body_dict(self, response_body_dict):
        versionner = self.get_specific_versionner()
        return versionner.response_body_dict(response_body_dict) if hasattr(versionner, "response_body_dict") else response_body_dict

    def set_request_args(self, request_args, request_kwargs):
        if "rest_helper_version" in request_kwargs:
            self.requested_version = request_kwargs["rest_helper_version"]
            request_kwargs.pop("rest_helper_version")

class UrlRootVersionner(BaseVersionner):
    @staticmethod
    def version_route(route):
        route.rule="/<rest_helper_version>"+route.rule


"""
This module contains functions that are geared toward serializing objects,
in particular JSON API objects.
"""

import decimal
from collections import Iterable
from rest_helpers.jsonapi_objects import Resource, Response, Link, JsonApiObject, Relationship

def to_jsonable(obj, no_empty_field=False):
    """
    This is a low level function to transform any object into a json
    serializable (jsonable) object based on its __dict__.

    Arguments:
        obj {any type} -- the object to be transformed.

    Keyword Arguments:
        no_empty_field {bool} -- if set to true, the empty field (empty
        string or None) will be removed from the resulting jsonable object
        (default: {False})

    Returns:
        dict -- A dictionary that can be used by json.dumps
    """
    if isinstance(obj, list):
        return [to_jsonable(r, no_empty_field) for r in obj]
    dic = obj if isinstance(obj, dict) else \
          obj.__dict__ if hasattr(obj, "__dict__") else \
          None

    if dic is None:
        if isinstance(obj, decimal.Decimal):
            str_rep = str(obj)
            return int(obj) if '.' not in str_rep else str_rep
        return obj

    return {str(k): to_jsonable(v, no_empty_field)for k, v in dic.items()if str(k)[0] != '_' and (not no_empty_field or v is not None and v != "")}

def response_to_jsonable(response, generate_self_links=True, id_only=False):
    """
    Transform a response object into a json serializable (jsonable) object that
    matches the jsonapi requirements.

    Arguments:
        resource {Response} -- The response to be serialized

    Keyword Arguments:
        generate_self_links {bool} -- If set to true "self" links will be added appropriately
        where they do not exist. (default: {True})

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the Response object.
    """
    assert isinstance(response, Response)

    # Data is a resource object (or a list of resource object,
    # hence it needs some special serialization logic)
    dic = response.__dict__.copy()
    dic.pop("data")
    return_value = to_jsonable(dic, no_empty_field=True)

    if response.data is not None:
        jsonable_data = resource_to_jsonable(response.data, generate_self_links)
        if id_only:
            jsonable_data = jsonable_data["id"] if not isinstance(jsonable_data, Iterable) else [x["id"] for x in jsonable_data]
        return_value["data"] = jsonable_data

    return return_value


def resource_to_jsonable(resource, generate_self_links=True):
    """
    Transform a resource object or a resource object list into
    a json serializable (jsonable) object that matches the jsonapi
    requirements.

    Arguments:
        resource {Resource|list<Resource>} -- The resource or list of resources
        to be serialized

    Keyword Arguments:
        generate_self_links {bool} -- If set to true "self" links will be added appropriately
        where they do not exist. (default: {True})

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the Resource object.
    """

    if isinstance(resource, list):
        return [resource_to_jsonable(x) for x in resource]

    assert isinstance(resource, Resource)

    json_resource = resource.to_primitive() if (hasattr(resource, "to_primitive") and callable(resource,to_primitive)) else to_jsonable(resource)
    special = ["id", "type", "relationships", "links", "meta"]
    for key in special:
        json_resource.pop(key, None)

    relationships = relationships_to_jsonable(
        resource.relationships, "{0}?json_path=/{1}".format(resource.id, "relationships"),
        generate_self_links)

    resource_links = resource.links
    if generate_self_links and "self" not in resource_links:
        resource_links = resource.links.copy()
        resource_links["self"] = Link(resource.id)

    links = links_to_jsonable(resource_links)
    return_value = {
        "id" : resource.id,
        "type" : resource.type,
        "relationships" : relationships,
        "links" : links,
        "meta" : resource.meta,
        "attributes" :json_resource
    }

    _remove_empty_fields(return_value)
    return return_value

def link_to_jsonable(link):
    """
    Transforms a json api link object into a dictionary that can be used by json.dumps.

    Arguments:
        link {Link} -- the link to be serialized.

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the Link object.
    """

    assert isinstance(link, Link)

    if link.meta is None:
        return  link.url
    else:
        return {
            "href": link.url,
            "meta": to_jsonable(link.meta)
        }

def links_to_jsonable(links):
    """
    Transform a json api Link object dictionary into a dictionaty that can be used
    by json dumps.

    Arguments:
        links {dict<Link>} -- the dictionary of Link objects to be serialized.

    Returns:
        dict --  a dictionary that can be used by json.dumps to serialize the dictionary of link
        objects.
    """

    if links is None:
        return None

    assert isinstance(links, dict)
    return {k: link_to_jsonable(v) for k, v in links.items()}

def jsonapiobject_to_jsonable(jsonapiobject):
    """
    Transforms a jsonapi json api objects into a dictionary that can be used by json dumps

    Arguments:
        jsonapiobject {JsonApiObject} -- The jsonapiobject to be serialized.

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the JsonApiObject object.
    """

    assert isinstance(jsonapiobject, JsonApiObject)
    return  to_jsonable(jsonapiobject, no_empty_field=True)


def relationship_to_jsonable(relationship, self_link=None):
    """
    Tranform a json api relationship object into a json serializable object that matches
    the json api specification.

    Arguments:
        relationship {Relationship} -- a relationship object to be serialized.

    Keyword Arguments:
        self_link {string} -- link to the relationship to be serialized. If not None, a link
        json api object will be created based on this value and added to the links of the
        relationship object to be serialized (default: {None}).

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the relationship object.
    """

    assert isinstance(relationship, Relationship)

    return_value = dict()

    links = relationship.links.copy() if relationship.links is not None else dict()
    if self_link is not None:
        links["self"] = Link(self_link)

    if any(links):
        return_value["links"] = links_to_jsonable(links)

    if relationship.data is not None:
        return_value["data"] = {"type": relationship.data.type, "id": relationship.data.id}

    return return_value

def relationships_to_jsonable(relationships, self_link_prefix=None, generate_self_link=False):
    """
    Tranform a dictionary of json api relationship object nto a json
    serializable object that matches the json api specification.

    Arguments:
        relationships {dict<Relationships>} -- a dict of
        relationship objects to be serialized.

    Keyword Arguments:
        self_link_prefix {string} -- prefix to be used as the link prefix when generate_self_link
        is set to true. (default: {None})
        generate_self_link {bool} -- when set to true, a self link will be autogenerated when
        serializing the relationship object (default: {False}).

    Returns:
        dict -- a dictionary that can be used by json.dumps to serialize the relationship
        dictionary.
    """

    if relationships is None:
        return None

    assert isinstance(relationships, dict)

    if generate_self_link:
        return {k: relationship_to_jsonable(v, "{0}/{1}".format(self_link_prefix, k))
                for k, v in relationships.items()}
    else:
        return {k: relationship_to_jsonable(v) for k, v in relationships.items()}

#region private

def _remove_empty_fields(dic):
    for key in [k for k, v in dic.items() if v is None or v == ""]:
        dic.pop(key)

#endregion



import yaml
import re
from rest_helpers import routes,jsonapi_objects, binding, validators, type_deserializers, routes

SWAGGER_AUGMENT_DEFAULT_KEY = "augment_default"
SWAGGER_EXTRA_DEFINITION_KEY = "extra_definition"
SWAGGER_DOCUMENTATION_KEY = "doc"
SWAGGER_PARAMETER_KEY = "parameters"
SWAGGER_JSONAPI_RESPONSE_DEFINITION_KEY = "jsonapi_response"

_swagger_object_classes = [jsonapi_objects.ErrorResponse, jsonapi_objects.SuccessResponse]

def get_swagger_service_description(source):
    """
    This function gets the full swagger spec for a given service.
    
    Arguments:
        source {dict|func|class} -- The source can be either a dictionary containing the high
                                    level description of the service or a func/class whose
                                    docstring contains the service high level description as a
                                    yaml blob under a 'Swagger doc:' line.
    
    Returns:
        dict -- the swagger description of the service.
    """

    default_service_description = {
        "swagger": "2.0",
        "info": {
            "description" : "A new service",
            "version" : "1.0.0",
            "title" : "New Service title",
            "contact" : {
                "email": "noreply@abc.com"
            }
        },
        "host": "service.domain.com",
        "basePath": "",
        "tags":[],
        "schemes": ["http"],
        "paths":get_swagger_paths(),
        "definitions":get_swagger_definitions()
    }

    if isinstance(source, dict):
        augment_dic = source
    else:
        augment_dic =  _get_swagger_part(source, SWAGGER_DOCUMENTATION_KEY)

    _update(default_service_description, augment_dic)

    return default_service_description


def swagger_object(cls):
    _swagger_object_classes.append(cls)
    return cls


def get_swagger_definitions():
    return_value = {}
    for swagger_obj in _swagger_object_classes:
        if issubclass(swagger_obj, jsonapi_objects.Resource):
            swagger_obj_type = _get_swagger_type(swagger_obj)

            jsonapi_response_def = _get_swagger_part(swagger_obj, SWAGGER_JSONAPI_RESPONSE_DEFINITION_KEY)
            if jsonapi_response_def is not None:
                inner_type = "{0}_data".format(swagger_obj_type)
                swagger_array_obj_type = "{0}_array".format(swagger_obj_type)
                return_value[swagger_obj_type] = {
                    "type": "object",
                    "allOf": [{"$ref": "#/definitions/response"}],
                    "properties":{
                        "data": {"$ref":"#/definitions/{0}".format(inner_type)}
                        }
                }
                return_value[swagger_array_obj_type] = {
                    "type": "object",
                    "allOf": [{"$ref": "#/definitions/response"}],
                    "properties":{
                        "data": {
                            "type":"array",
                            "items":{
                                "$ref": "#/definitions/{0}".format(inner_type)
                            }
                        }
                    }
                }

                return_value[inner_type] = {
                    "type": "object",
                    "properties":{
                        "attributes":{
                            "type": "object",
                            "properties": jsonapi_response_def
                        }
                    }
                }
            else:
                return_value[swagger_obj_type] = _get_swagger_part(swagger_obj, SWAGGER_DOCUMENTATION_KEY)


        extra_definitions = _get_swagger_part(swagger_obj, SWAGGER_EXTRA_DEFINITION_KEY)
        _update(return_value, extra_definitions)
    return return_value


def get_swagger_paths():
    return_value = {}
    for route in routes._swagger_routes:
        existing_doc = _get_swagger_part(route.view_function, SWAGGER_DOCUMENTATION_KEY)
        key = "/"+route.rule.replace("<","{").replace(">","}").strip(" /")
        if key not in return_value:
            return_value[key] = {}

        _update(return_value[key], _get_default_swagger_path(route))
        _update(return_value[key][next(iter(return_value[key].keys()))], existing_doc)

        augment_dic = _get_swagger_part(route.view_function, SWAGGER_AUGMENT_DEFAULT_KEY)
        if augment_dic is not None:
            _update(return_value[key], { route.options["methods"][0].lower() : augment_dic})
    return return_value


def _get_swagger_type(resource_class):
    return resource_class.resource_type.replace("/","")


def _get_swagger_part(function_or_class, key="doc"):
    doc = function_or_class.__doc__

    swagger_key = "Swagger {0}:".format(key)
    if doc is not None and swagger_key in doc:
        swagger_text = doc.split(swagger_key)[1]
        swagger_text = swagger_text.split("end swagger")[0]
        return yaml.load(swagger_text)

    return None

#region default paths
def _get_default_swagger_path(route):
    return_value = None

    if type(route) == routes.get_resource_route:
        return_value = _get_default_swagger_path_for_get_route(route)
    elif type(route) == routes.get_all_resources_route:
        return_value = _get_default_swagger_path_for_get_all_route(route)
    elif type(route) == routes.put_resource_route:
        return_value = _get_default_swagger_path_for_put_route(route)
    elif type(route) == routes.operation_resource_route:
        return_value = _get_default_swagger_path_for_op_route(route)
    elif type(route) == routes.group_operation_resource_route:
        return_value = _get_default_swagger_path_for_group_op_route(route)
    elif type(route) == routes.patch_resource_route:
        return_value = _get_default_swagger_path_for_patch_resource_route(route)
    elif type(route) == routes.delete_resource_route:
        return_value = _get_default_swagger_path_for_delete_resource_route(route)
    return return_value


def _get_default_swagger_path_for_get_route(route):
    return_value = {
        "get":{
            "tags":[route.resource_class.resource_type],
            "summary": "Get by name".format(type=route.resource_class.resource_type),
            "description":"",
            "operationId": "get{type}".format(type="_".join((unpluralize(x) for x in route.resource_class.resource_type.split("/")))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route),
            "responses":{
                "200":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "404":
                {
                    "description": "Not found: the requested {type} could not be found.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }

    return return_value


def _get_default_swagger_path_for_get_all_route(route):
    return_value = {
        "get":{
            "tags":[route.resource_class.resource_type],
            "summary": "Get all".format(type=route.resource_class.resource_type),
            "description":"",
            "operationId": "get_all{type}".format(type="_".join(x for x in route.resource_class.resource_type.split("/"))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route),
            "responses":{
                "200":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}_array".format(type=_get_swagger_type(route.resource_class))
                    }
                }
            }
        }
    }

    return return_value

def _get_default_swagger_path_for_put_route(route):
    parameters = [p for p in  _get_parameters(route) if p.get("in") != "body"]
    parameters.append({
                "in": "body",
                "name": "body",
                "description": "Object to be created.",
                "required": True,
                "schema":{ "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)) }
            })
            
    return_value = {
        "put":{
            "tags":[route.resource_class.resource_type],
            "summary": "Create or edit".format(type=route.resource_class.resource_type),
            "description":"",
            "operationId": "put{type}".format(type="_".join((unpluralize(x) for x in route.resource_class.resource_type.split("/")))),
            "produces": ["application/json"],
            "consumes": ["application/json"],
            "parameters": parameters,
            "responses":{
                "201":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "400":
                {
                    "description": "The body of the request is improperly formed.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }

    return return_value

def _get_default_swagger_path_for_patch_resource_route(route):
    return_value = {
        "patch":{
            "tags":[route.resource_class.resource_type],
            "summary": "Patch resource".format(type=route.resource_class.resource_type),
            "description":"",
            "operationId": "patch{type}".format(type="_".join((unpluralize(x) for x in route.resource_class.resource_type.split("/")))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route),
            "responses":{
                "201":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "400":
                {
                    "description": "The body of the request is improperly formed.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }
    return return_value

def _get_default_swagger_path_for_op_route(route):
    return_value = {
        "post":{
            "tags":[route.resource_class.resource_type],
            "summary": "Operation {operation_name}".format(operation_name=route.operation_name),
            "description":"",
            "operationId": "{operation_name}{type}".format(operation_name=route.operation_name, type="_".join((unpluralize(x) for x in route.resource_class.resource_type.split("/")))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route, route.operation_name),
            "responses":{
                "200":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "400":
                {
                    "description": "The body of the request is improperly formed.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }

    return return_value

def _get_default_swagger_path_for_group_op_route(route):
    return_value = {
        "post":{
            "tags":[route.resource_class.resource_type],
            "summary": "Group operation {operation_name}".format(operation_name=route.operation_name),
            "description":"",
            "operationId": "{operation_name}_all{type}".format(operation_name=route.operation_name, type="_".join(route.resource_class.resource_type.split("/"))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route, route.operation_name),
            "responses":{
                "200":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "400":
                {
                    "description": "The body of the request is improperly formed.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }

    return return_value

def _get_default_swagger_path_for_delete_resource_route(route):
    return_value = {
        "delete":{
            "tags":[route.resource_class.resource_type],
            "summary": "Delete ".format(type=route.resource_class.resource_type),
            "description":"",
            "operationId": "delete{type}".format(type="_".join((unpluralize(x) for x in route.resource_class.resource_type.split("/")))),
            "produces": ["application/json"],
            "parameters": _get_parameters(route),
            "responses":{
                "200":
                {
                    "description": "successful operation",
                    "schema":
                    {
                        "$ref": "#/definitions/{type}".format(type=_get_swagger_type(route.resource_class)),
                    }
                },
                "404":
                {
                    "description": "Not found: the requested {type} could not be found.".format(type=route.resource_class.resource_type),
                    "schema":
                    {
                        "$ref": "#/definitions/error",
                    }
                }
            }
        }
    }

    return return_value
#endregion


def _get_parameters(route, operation_name=None):
    parameter_names = re.finditer("<(\w+)>", route.rule)

    action = "to get." if operation_name is None else "to apply {operation_name} on".format(operation_name=operation_name)
    parameters = [
    {
        "name":p.group(1),
        "in":"path",
        "description":"Name of the {type} {action}".format(type=route.resource_class.resource_type, action=action) if isinstance(route, routes.base_resource_route) else "",
        "required":True,
        "type":"string"
    } for p in parameter_names]

    version_parameter = next((p for p in parameters if p["name"]=="rest_helper_version"), None)
    if version_parameter is not None:
        version_parameter["description"] = "Version of the api to use."
        version_parameter["enum"] = sorted(list(route.versionner().supported_versions.keys()), reverse=True)

    if route.id in binding._input_decorators:
        swagger_parameter_dict = _get_swagger_part(route.real_view_function, SWAGGER_PARAMETER_KEY) or {}

        def merge(x,y):
            if y is None:
                return None
            x.update(y)
            return x

        query_decorators = (p for p in binding._input_decorators[route.id] if isinstance(p, binding.from_query_string))
        query_parameters = [
        merge({
            "name":p.query_field,
            "in":"query",
            "required": not p.has_default,
            "type": _get_parameter_type(p.type),
        }, swagger_parameter_dict.get(p.query_field,{})) for p in query_decorators]
        parameters += [x for x in query_parameters if x is not None]

        header_decorators = (p for p in binding._input_decorators[route.id] if isinstance(p, binding.from_header))
        header_parameters = [
        merge({
            "name":p.header_field,
            "in":"header",
            "required": not p.has_default,
            "type": _get_parameter_type(p.type),
        }, swagger_parameter_dict.get(p.header_field,{})) for p in header_decorators]

        parameters += [x for x in header_parameters if x is not None]

        body_parameters = [p for p in binding._input_decorators[route.id] if isinstance(p, binding.from_json_body) or isinstance(p, binding.field_from_json_body)]
        if any(body_parameters) and  ("body" not in swagger_parameter_dict or swagger_parameter_dict["body"] is not None):
            body_parameters = [
            merge({
                "name":"body",
                "in":"body",
                "required": any(not p.has_default for p in body_parameters),
                "schema": {"type": "object"}
                },
                swagger_parameter_dict.get("body", {}))
            ]

            if body_parameters is not None:
                parameters += body_parameters
        """
        body_decorators = (p for p in binding._input_decorators[route.id] if isinstance(p, binding.from_json_body))
        body_parameters = [
        {
            "name":"body",
            "in":"body",
            "required": not p.has_default,
            "type": "object"
        } for p in body_decorators]

        parameters += body_parameters

        field_from_body_decorators = (p for p in binding._input_decorators[route.id] if isinstance(p, binding.field_from_json_body))
        field_from_body_parameters = [
        {
            "name":p.json_field,
            "in":"body",
            "required": not p.has_default,
            "type": _get_parameter_type(p.type)
        } for p in field_from_body_decorators]

        parameters += field_from_body_parameters
        """

    if type(route) == routes.get_all_resources_route and route.page_size is not None:
        parameters.append({
            "name":"page",
            "in":"query",
            "description":"The page number.",
            "required":False,
            "type":"integer"
        })

    return parameters


def _get_parameter_type(p_type):
    if p_type == bool:
        return "boolean"
    elif p_type == int:
        return "integer"
    else:
        return "string"


def _update(d, u):
    """
    Updates dictionary d with values from u recursively.
    when updating a lidt, the last two element of the key
    are taken in account :
      - >+: add all elements of the list to the list
      - >-: remove all elements of the list from the list
      - else replace the list

    Arguments:
        d {dict} -- the dictionary to be updated.
        u {dict} -- the dictionary containing the update.
    
    Returns:
        dict -- the updated dictionary.
    """
    if u is None:
        return d

    for _k, v in u.items():
        # The following block allows us to add a modifer
        # in the last two character of a key ">X"
        if isinstance(_k, str) and len(_k)>2 and _k[-2] == ">":
            k=_k[0:-2]
            modifier=_k[-1]
        else:
            k = _k
            modifier=None

        if (k in d or isinstance(d,list) and k<len(d)) and isinstance(v, dict):
            _update(d[k], v)
        elif k in d and (isinstance(v,list) or modifier=='-') and isinstance(d[k],list):
            if modifier == '+':
                d[k] += v
            elif modifier == '-':
                for el in v:
                    if re.match("\[[0-9]+\]", str(el)):
                        index = int(el[1:-1])
                        del d[k][index]
                    else:
                        d[k].remove(el)
            else:
                d[k] = v
        else:
            d[k] = v
    return d

def unpluralize(word):
    return word[:-3] + "y" if word [-3:] == "ies" else word[:-1]


from rest_helpers import type_deserializers


class TestTypeDeserializers(object):
    def test_bool_deserializer(self):
        assert type_deserializers.bool_deserializer("")
        assert type_deserializers.bool_deserializer("True")
        assert type_deserializers.bool_deserializer("true")

        assert not type_deserializers.bool_deserializer("false")
        assert not type_deserializers.bool_deserializer("False")
        assert not type_deserializers.bool_deserializer(None)

    def test_int_deserializer(self):
        assert type_deserializers.int_deserializer("3") == 3


#!/usr/bin/env python3

import os
import logging
import sys
try:
    import multiprocessing
except:
    pass
# nose requires multiprocessing and logging to be initialized before the setup
# call, or we'll get a spurious crash on exit.
from setuptools import setup, find_packages
from setuptools.dist import Distribution

is_release = False
if "--release" in sys.argv:
    is_release = True
    sys.argv.remove("--release")

base = os.path.dirname(os.path.abspath(__file__))


def read(fname):
    '''Utility function to read the README file.'''
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# figure out what the install will need
install_requires = [
    "setuptools >=0.5",
    "flask",
    "pyyaml",
    "requests",
    "schematics",
    "python-dateutil",
    "requests-futures",
    "httpretty",
    "aiohttp>=2.3.0",
    "aiotask_context",
    "pytest-aiohttp",
    "pytest-asyncio",
    'cryptography',
    'python-jose[cryptography]',
    'jinja2'
]

setup(
    name="rest-helpers",
    setup_requires=["vcver"],
    vcver={
        "is_release": is_release,
        "path": base
    },
    url="https://github.com/WillFr/rest-helpers",
    author="Guillaume Grosbois",
    author_email="grosbois.guillaume@gmail.com",
    description="A set of method to help creating rest services",
    packages=find_packages(),
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Operating System :: OS Independent"],
    package_data={"rest_helpers": ["templates/swagger-ui.html"]},
    install_requires=install_requires,
    include_package_data=True,
    tests_require=[  "mock >=0.7.2",
                     "coverage",
                     "httpretty",
                     "httmock",
                    "pytest-aiohttp",
                    "pytest-cov"] + install_requires
    )



import sys
import json
import asyncio
import aiotask_context
from aiohttp import web,web_request
from multidict import MultiDict
from rest_helpers.framework_adapter import BaseFrameworkAdapter

def aiohttp_adapter_builder(*args, **kwargs):
    if any(args) and isinstance(args[0], web.Application):
        adapter = AioHttpFrameworkAdapter(args[0])
        args = args[1:]
    else:
        adapter = AioHttpFrameworkAdapter()

    return adapter, args, kwargs


class AioHttpFrameworkAdapter(BaseFrameworkAdapter):
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            app.loop.set_task_factory(aiotask_context.task_factory)

    def is_in_test(self):
        # TODO
        return False #not has_request_context() or request_context.top.app.config["TESTING"] is True

    async def get_current_request_body(self):
        return await self.get_rest_helper_request_context().request.text()

    def get_current_request_query_string_args(self):
        query_multidict=self.get_rest_helper_request_context().request.query
        return dict((k,query_multidict.getall(k)) for k in query_multidict.keys())


    def get_current_request_query_string(self):
        return self.get_rest_helper_request_context().request.query_string

    def get_current_request_headers_dict(self):
        return self.get_rest_helper_request_context().request.headers

    def get_current_request_path(self):
        return self.get_rest_helper_request_context().request.path

    def attach_rest_helper_request_context(self, context):
        current_task = asyncio.Task.current_task()
        if not hasattr(current_task, "context"):
            current_task.context = {}

        aiotask_context.set(key="rest_helper_context", value=context)

    def get_current_request_full_path(self):
        return self.get_rest_helper_request_context().request.path_qs

    def get_current_request_url(self):
        return self.get_rest_helper_request_context().request.url

    def add_url_rule(self, route, func):
        self.app.router.add_route(route.options.get("method",["GET"])[0], route.rule.replace("<","{").replace(">","}"), func)

    def get_rest_helper_request_context(self):
        #return asyncio.Task.current_task().rest_helper_context  if hasattr(asyncio.Task.current_task(), "rest_helper_context") else None
        return aiotask_context.get(key="rest_helper_context", default=None)

    def set_request_args(self, args):
        if len(args)>0 and isinstance(args[0], web_request.Request):
            self.get_rest_helper_request_context().request = args[0]
            return args[1:] if len(args)>1 else []
        return args or []

    def set_request_kwargs(self, kwargs):
        kwargs.update(self.get_rest_helper_request_context().request.match_info)
        return kwargs

    def get_current_request_method(self):
        #TODO :
        raise NotImplementedError()

    def get_current_request_headers(self):
        return {}

    def make_json_response(self, obj, status = 200, headers=None, title=None):
        headers = headers or {}
        json_content = json.dumps(obj)
        return web.Response(body=json_content, headers = MultiDict(headers), status=status, reason=title)

def add_default_swagger_routes(app, source):
    import rest_helpers.routes as native_routes
    native_routes.add_default_swagger_routes(AioHttpFrameworkAdapter(app), source)




import json
import pytest

from mock import patch, Mock, MagicMock
from rest_helpers import responses, framework_adapter, rest_helper_context, rest_exceptions, binding
from rest_helpers.jsonapi_objects import Resource

# TODO test success response with non Resource object

@pytest.fixture
def test_adapter(request):
    adapter = MagicMock(wraps=framework_adapter.BaseFrameworkAdapter())
    adapter.make_json_response = lambda  obj,status=200, headers=None, title=None: (obj, status, headers, title)
    adapter.get_current_request_path = lambda: "/"
    adapter.get_current_request_query_string_args = lambda:{}
    adapter.get_current_request_headers_dict = lambda: {}
    adapter.get_current_request_query_string = lambda:b""
    rh_context = rest_helper_context.RestHelperContext()
    adapter.get_rest_helper_request_context = lambda: rh_context
    return adapter

def test_bad_request(test_adapter):
    ex = Exception("test exception")
    resp = responses.bad_request(test_adapter, ex)
    assert resp[1]== 400
    assert str(ex) in str(resp[0])
    assert "Bad request" in str(resp[0])

def test_not_found(test_adapter):
    detail_string = "This is the details about the issue."
    resp = responses.not_found(test_adapter, detail_string)
    assert detail_string in str(resp[0])

def test_not_found_with_details(test_adapter):
    resp = responses.not_found(test_adapter)
    assert resp[1]== 404
    assert "Not found" in str(resp[0])
    assert "The requested resource does not exist" in str(resp[0])

@pytest.mark.parametrize("headers", [
    {"X-Unique-ID": "this is the unique request id"},
    {}
])
def test_internal_server_error(test_adapter, headers):
    test_adapter.get_current_request_headers_dict = lambda: headers
    resp = responses.internal_server_error(test_adapter, Exception("test exception"), "stack trace")
    assert resp[1]== 500
    assert "Internal server error" in str(resp[0])
    assert "test exception" in str(resp[0])
    assert "stack trace" in str(resp[0])
    assert resp[3] == 'Internal server error'

    if headers !=  {}:
        assert "this is the unique request id" in str(resp[0])


def test_ok(test_adapter):
    resp = responses.ok(test_adapter, Resource("type", "name"))
    assert resp[1]== 200

def test_ok_id_only(test_adapter):
    resp = responses.ok(test_adapter, [Resource("name1", "type"), Resource("name2", "type"), Resource("name3", "type")], id_only=True)
    assert resp[1]== 200
    assert all("/type/name" in x for x in resp[0]["data"])

def test_created(test_adapter):
    resp = responses.created(test_adapter, Resource("type", "name"))
    assert resp[1]== 201

def test_accepted(test_adapter):
    resp = responses.accepted(test_adapter, Resource("type", "name"))
    assert resp[1]== 202

    resp = responses.accepted(test_adapter)
    assert resp[1]== 202

def test_ok_with_json_path(test_adapter):
    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/name"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == "test_name"
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/list"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == ["1",2,"3"]
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/list/1"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == 2
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/dic/entry_1"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == "value 1"
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/dic/list/*/x"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == [1,2,3]
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/list/3"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[1]== 400

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/list/a"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[1]== 400

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/a"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[1]== 400

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/dic/list/*:>b~=(2|3)/b"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[0]["data"] == ["2","3"]
    assert resp[1]== 200

    test_adapter.get_current_request_query_string_args = lambda:{"json_path":["/data/attributes/list/*:sfds"]}
    resp = responses.ok(test_adapter, _TestClass())
    assert resp[1]== 400

def _verify_ok_page_1(resp):
    assert resp[1]== 200
    data = resp[0]
    assert set(x["id"] for x in data["data"]) == set([1,2])
    assert "page=2" in data["links"]["next"]["url"]
    assert data["meta"]["total_pages"]==3
    assert "last" not in data["links"]

def _verify_ok_page_2(resp):
    assert resp[1]== 200
    data = resp[0]
    assert set(x["id"] for x in resp[0]["data"]) == set([3,4])
    assert "page=3&x=3" in data["links"]["next"]["url"]
    assert "page=1&x=3" in data["links"]["last"]["url"]

def _verify_ok_page_3(resp):
    assert resp[1]== 200
    data = resp[0]
    assert set(x["id"] for x in resp[0]["data"]) == set([5])
    assert "page=2" in data["links"]["last"]["url"]
    assert "next" not in data["links"]
    return

def test_ok_with_explicit_page_size(test_adapter):
    resp_list = [_TestClass(1), _TestClass(2), _TestClass(3), _TestClass(4), _TestClass(5)]

    test_adapter.get_current_request_query_string_args = lambda:{}
    test_adapter.get_current_request_query_string = lambda:b""
    resp = responses.ok(test_adapter, resp_list, page_size=2)
    _verify_ok_page_1(resp)

    test_adapter.get_current_request_query_string_args = lambda:{"page":["2"], "x":["3"]}
    test_adapter.get_current_request_query_string = lambda:b"page=2&x=3"
    resp = responses.ok(test_adapter, resp_list, page_size=2)
    _verify_ok_page_2(resp)

    test_adapter.get_current_request_query_string_args = lambda:{"page":["3"]}
    test_adapter.get_current_request_query_string = lambda:b"page=3"
    resp = responses.ok(test_adapter, resp_list, page_size=2)
    _verify_ok_page_3(resp)

def test_ok_with_context_page_size(test_adapter):
    resp_list = [_TestClass(1), _TestClass(2), _TestClass(3), _TestClass(4), _TestClass(5)]
    test_adapter.get_current_request_query_string_args = lambda:{}
    test_adapter.get_current_request_query_string = lambda:b""
    test_adapter.get_rest_helper_request_context().page_size = 2
    resp = responses.ok(test_adapter, resp_list)
    _verify_ok_page_1(resp)

    test_adapter.get_current_request_query_string_args = lambda:{"page":["2"], "x":["3"]}
    test_adapter.get_current_request_query_string = lambda:b"page=2&x=3"
    test_adapter.get_rest_helper_request_context().page_size = lambda : 2
    resp = responses.ok(test_adapter, resp_list)
    _verify_ok_page_2(resp)


    test_adapter.get_current_request_query_string_args = lambda:{"page":["3"]}
    test_adapter.get_current_request_query_string = lambda:b"page=3"
    test_adapter.get_rest_helper_request_context().page_size = lambda : 4-2
    resp = responses.ok(test_adapter, resp_list)
    _verify_ok_page_3(resp)

def test_ok_with_query_string_page_size(test_adapter):
    resp_list = [_TestClass(1), _TestClass(2), _TestClass(3), _TestClass(4), _TestClass(5)]
    test_adapter.get_current_request_query_string_args = lambda:{"page_size":"2"}
    test_adapter.get_current_request_query_string = lambda:b"page_size=2"
    test_adapter.get_rest_helper_request_context().page_size = None
    resp = responses.ok(test_adapter, resp_list)
    _verify_ok_page_1(resp)

    test_adapter.get_current_request_query_string_args = lambda:{"page_size":"2", "page":["2"], "x":["3"]}
    test_adapter.get_current_request_query_string = lambda:b"page_size=2&page=2&x=3"
    test_adapter.get_rest_helper_request_context().page_size = None
    resp = responses.ok(test_adapter, resp_list)
    _verify_ok_page_2(resp)

class _TestClass(Resource):
    def __init__(self, id = 1):
        super(_TestClass, self).__init__("test_name", "test_type")
        self.id=id
        self.list = ["1",2,"3"]
        self.dic = {
            "entry_1": "value 1",
            "list": [{"x":1,"b":"2"},{"x":2,"b":"3"},{"x":3,"b":"4"}]
        }

def test_parse_filter():
    filter = responses._parse_filter("*:>a>b==3")
    test_data = [
        {"a":{"b":2}, "c":4},
        {"a":{"b":3}, "c":4},
        {"a":{"b":3, "c":4}},
    ]

    filtered = [x for x in test_data if filter(x)]
    assert filtered == [{'a': {'b': 3}, 'c': 4}, {'a': {'b': 3, 'c': 4}}]

@pytest.mark.parametrize("exception, expected_code, expected_string, expected_title", [
    (rest_exceptions.InvalidDataException("The data X is not valid"), 400, "Client error: The data X is not valid", "Bad request"),
    (binding.MissingFieldException("The field XYZ is missing"), 400, "Client error: The field XYZ is missing", "Bad request"),
    (rest_exceptions.UnauthorizedException("You are not authorized"), 401, "Your authentication was not successful.", "Unauthorized"),
    (rest_exceptions.ForbiddenException("This is forbidden"), 403, "You are not authorized to access the requested resources or perform the requested operation.", "Forbidden")
])
def test_exception_handler(test_adapter, exception, expected_code, expected_string, expected_title):
    result_json, status_code, headers, title = responses.base_exception_handler(test_adapter, exception)

    assert status_code == expected_code
    assert title == expected_title
    assert expected_string in result_json["error"]["detail"]


import sys
import asyncio
import inspect
import functools
from flask import jsonify,request,has_request_context, Blueprint, jsonify, Flask
from flask.globals import _request_ctx_stack as request_context
from multiprocessing.pool import ThreadPool
from rest_helpers.framework_adapter import BaseFrameworkAdapter

def flask_adapter_builder(*args, **kwargs):
    if any(args) and isinstance(args[0], Blueprint):
        adapter = FlaskFrameworkAdapter(args[0])
        args = args[1:]
    else:
        adapter = FlaskFrameworkAdapter()

    return adapter, args, kwargs


_async_handled = False
_async_map = {}
def handle_async_route(loop=None):
    """
    This functions monkey patches the Flask class to transform asynchrnous view functions into
    synchronous ones.

    The async handling should be called only once (which is guarded by the `async_handled` global variable)
    Each route view function will be wrapped in a sync_function only once and this mapping will be stored
    in the `_async_map` variable
    """

    global _async_handled
    if _async_handled is True:
        return
    _async_handled = True

    original = Flask.add_url_rule
    if loop is None:
        try:
            loop = asyncio.get_event_loop()
        except: #pragma: no cover
            loop = asyncio.new_event_loop()

    def replacement(self, rule, endpoint=None, view_func=None, **options):
        global _async_map
        sync_function = _async_map.get(view_func)
        if sync_function is None:
            if view_func is not None :
                def sync_function(*args, **kwargs):
                    result = view_func(*args, **kwargs)
                    return loop.run_until_complete(result) if inspect.iscoroutine(result) else result

                functools.update_wrapper(sync_function, view_func)
                _async_map[view_func] = sync_function
            else:
                sync_function = view_func

        return original(self, rule, endpoint=endpoint or "sync-{}".format(view_func.__name__), view_func=sync_function, **options)

    Flask.add_url_rule = replacement

class FlaskFrameworkAdapter(BaseFrameworkAdapter):
    def __init__(self, blueprint=None):
        self.blueprint = blueprint
        handle_async_route()
        try:
            self.loop = asyncio.get_event_loop()
        except: #pragma: no cover
            self.loop = asyncio.new_event_loop()

    def is_in_test(self):
        return not has_request_context() or request_context.top.app.config["TESTING"] is True

    async def get_current_request_body(self):
        return request.data.decode()

    def get_current_request_query_string_args(self):
        return dict(request.args.lists())

    def get_current_request_query_string(self):
        return request.query_string

    def get_current_request_headers_dict(self):
        return request.headers

    def get_current_request_path(self):
        return request.path

    def attach_rest_helper_request_context(self, context):
        if not hasattr(request, "rest_helper_context"):
            request.rest_helper_context = context

    def get_current_request_full_path(self):
        return request.full_path

    def get_current_request_url(self):
        return request.url

    def add_url_rule(self, route, func):
        endpoint = route.options.pop("endpoint", route.real_view_function.__name__)
        route.rule = (self.blueprint.url_prefix or "") + route.rule
        self.blueprint.add_url_rule(route.rule, endpoint, func, **route.options)

    def get_rest_helper_request_context(self):
        return request.rest_helper_context if hasattr(request, "rest_helper_context") else None

    def get_current_request_method(self):
        return request.method

    def get_current_request_headers(self):
        return {}

    def make_json_response(self, obj, status = 200, headers=None, title=None):
        headers = headers or {}
        response = jsonify(obj)
        if title is not None:
            response._status = str(status)+" "+title
            response._status_code = status
        else:
            response.status_code = status

        for k,v in headers.items():
            response.headers[k] = v
        return response

def add_default_swagger_routes(app, source, **kwargs):
    swagger_ui = Blueprint('swagger_ui', 'swagger_ui', url_prefix='')
    import rest_helpers.routes as native_routes
    native_routes.add_default_swagger_routes(FlaskFrameworkAdapter(swagger_ui), source, **kwargs)
    app.register_blueprint(swagger_ui)



import pytest
import asyncio
from collections import defaultdict

@pytest.fixture
def counter(request):
    return defaultdict(int)

@pytest.fixture
def loop(event_loop):
     """Ensure usable event loop for everyone.

     If you comment this fixture out, default pytest-aiohttp one is used
     and things start failing (when redis pool is in place).
     """
     return event_loop

@pytest.fixture
def make_coro():
    def inner(obj):
        mock_return_value = asyncio.Future()
        mock_return_value.set_result(obj)
        return mock_return_value

    return inner